# -*- coding: utf-8 -*-
# pyknic/lib/fastapi/lobby.py
#
# Copyright (C) 2025-2026 the pyknic authors and contributors
# <see AUTHORS file>
#
# This file is part of pyknic.
#
# pyknic is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# pyknic is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with pyknic.  If not, see <http://www.gnu.org/licenses/>.

import typing

from abc import ABCMeta, abstractmethod

import pydantic

from pyknic.lib.registry import APIRegistry, register_api
from pyknic.lib.fastapi.models.lobby import LobbyCommandResult


class LobbyCommandError(Exception):
    """This exception is raised for command execution errors."""
    pass


class LobbyCommandHandler(metaclass=ABCMeta):
    """Prototype for a lobby command implementation."""

    def __init__(self, args: pydantic.BaseModel):
        if not isinstance(args, self.command_model()):
            raise TypeError(
                f'Invalid argument type for the "{self.command_name()}" command. '
                f'The "{self.command_model().__class__}" was expected, but got {args.__class__}'
            )

        self._args = args

    @classmethod
    @abstractmethod
    def command_name(cls) -> str:
        """Return unique name of this command."""
        raise NotImplementedError('This method is abstract')

    @classmethod
    @abstractmethod
    def command_model(cls) -> typing.Type[pydantic.BaseModel]:
        """Return a model that represents command's arguments.
        """
        raise NotImplementedError('This method is abstract')

    @classmethod
    def prepare_command(cls, args: pydantic.BaseModel) -> 'LobbyCommandHandler':
        """Return instance of this object that is able to call the :meth:`.LobbyCommandHandler.exec` method.

        :param args: command arguments
        """
        return cls(args)

    @abstractmethod
    async def exec(self) -> LobbyCommandResult:
        """ Execute the command and return its result.
        """
        raise NotImplementedError('This method is abstract')


__default_lobby_commands_registry__ = APIRegistry()  # default registry for all commands


def register_lobby_command(
    registry: APIRegistry | None = None,
) -> typing.Callable[..., typing.Any]:
    """This decorator help to register commands with the given registry."""

    if registry is None:
        registry = __default_lobby_commands_registry__

    return register_api(
        registry=registry,
        api_id=lambda x: x.command_name(),
        callable_api_id=True
    )



# -*- coding: utf-8 -*-
# pyknic/lib/datalog/datalog_py.py
#
# Copyright (C) 2024 the pyknic authors and contributors
# <see AUTHORS file>
#
# This file is part of pyknic.
#
# pyknic is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# pyknic is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with pyknic.  If not, see <http://www.gnu.org/licenses/>.

import typing

from pyknic.lib.verify import verify_value

from pyknic.lib.datalog.proto import DatalogProto
from pyknic.lib.thread import CriticalResource


class DatalogPy(DatalogProto, CriticalResource):
    """ The :class:`.DatalogProto` implementation in Python with locks
    """

    def __init__(self, cr_timeout: typing.Union[int, float, None] = None):
        """ Create a new log

        :param cr_timeout: timeout for a lock acquiring
        """
        DatalogProto.__init__(self)
        CriticalResource.__init__(self, cr_timeout)
        self.__log: typing.List[typing.Any] = list()

    @CriticalResource.critical_section
    def append(self, record: typing.Any) -> None:
        """ The :meth:`.DatalogPy.append` method implementation
        """
        self.__log.append(record)
        self.emit(DatalogProto.new_entry, record)

    def iterate(self, reverse: bool = False) -> typing.Generator[typing.Any, None, None]:
        """ The :meth:`.DatalogPy.iterate` method implementation
        """

        with self.critical_context():
            log_copy = self.__log.copy()

        if reverse:
            log_copy.reverse()

        for i in log_copy:
            yield i

    @verify_value(min_length=lambda x: x >= 0)
    @CriticalResource.critical_section
    def truncate(self, min_length: int) -> None:
        """ The :meth:`.DatalogPy.truncate` method implementation
        """

        if min_length == 0:
            self.__log.clear()
        else:
            self.__log = self.__log[-min_length:]

    def find(
        self, filter_fn: typing.Callable[[typing.Any], bool], reverse: bool = False
    ) -> typing.Optional[typing.Any]:
        """ The :meth:`.DatalogPy.search` method implementation
        """
        try:
            return next(filter(filter_fn, self.iterate(reverse=reverse)))
        except StopIteration:
            return None



# -*- coding: utf-8 -*-
# pyknic/lib/tasks/proto.py
#
# Copyright (C) 2016-2024 the pyknic authors and contributors
# <see AUTHORS file>
#
# This file is part of pyknic.
#
# pyknic is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# pyknic is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with pyknic.  If not, see <http://www.gnu.org/licenses/>.

import enum
import typing

from abc import ABCMeta, abstractmethod
from dataclasses import dataclass

from pyknic.lib.x_mansion import CapabilitiesAndSignals
from pyknic.lib.signals.proto import Signal
from pyknic.lib.signals.source import SignalSource
from pyknic.lib.capability import capability


class TaskStartError(Exception):
    """ This exception is raised when there is an error in starting a task
    """
    pass


class TaskStopError(Exception):
    """ This exception is raised when there is an error in stopping a task
    """
    pass


class NoSuchTaskError(Exception):
    """ This exception is raised when there is no requested task (it may be a request to start unknown task or
    request to stop already stopped task)
    """
    pass


@dataclass
class TaskResult:
    """ This class is used along with a completion signal defining a result of a completed task. In order to check
    whether a task was completed successfully the 'exception' property should be checked

    :note: the 'result' property may be not the same as the result from the original :meth:`TaskProto.start`
    method call
    """
    result: typing.Any = None                         # a result of completed record (if any)
    exception: typing.Optional[BaseException] = None  # an exception raised within a task (if any)


class TaskProto(CapabilitiesAndSignals):
    """ Basic task prototype. Derived classes must implement the only thing - :meth:`TaskProto.start`
    """

    task_started = Signal()              # a task started
    task_completed = Signal(TaskResult)  # a task completed

    @abstractmethod
    def start(self) -> None:
        """ Start a task
        """
        raise NotImplementedError('This method is abstract')

    @capability
    def stop(self) -> None:
        """ Try to stop this task gracefully.

        :raise NotImplementedError: if this task can not be stopped
        """
        raise NotImplementedError('The "stop" method is not supported')

    @capability
    def terminate(self) -> None:
        """ Try to stop this task at all costs

        :raise NotImplementedError: if this task can not be terminated
        """
        raise NotImplementedError('The "terminate" method is not supported')

    def task_name(self) -> typing.Optional[str]:
        """ Return a task name that optionally identifies a task
        """
        return None


@enum.unique
class ScheduledTaskPostponePolicy(enum.Enum):
    """ This is a policy that describes what should be done with a task if a scheduler won't be able to run
    it (like if the scheduler's limit of running tasks is reached).
    """
    wait = enum.auto()        # will postpone the task to execute it later
    drop = enum.auto()        # drop this task if it can't be executed at the moment
    keep_first = enum.auto()  # if there are postponed tasks, then drop this task
    keep_last = enum.auto()   # if there are postponed tasks, drop them and keep this task


class ScheduleRecordProto(metaclass=ABCMeta):
    """ This class describes a single request that scheduler (:class:`.SchedulerProto`) should process
    (should start). It has a :class:`.ScheduledTaskProto` task to be started and postpone policy
    (:class:`.TaskPostponePolicy`)

    Postpone policy is a recommendation for a scheduler and a scheduler can omit it if (for example) a scheduler queue
    is full. 'ScheduledTaskPostponePolicy.keep_first' and 'ScheduledTaskPostponePolicy.keep_last' postpone policies
    (so as simultaneous policy) will be applied to this task and to tasks with the same group id (if it was set).
    """

    @abstractmethod
    def task(self) -> TaskProto:
        """ Return a task that should be started
        """
        raise NotImplementedError('This method is abstract')

    @abstractmethod
    def source(self) -> 'ScheduleSourceProto':
        """ Return origin of this record
        """
        raise NotImplementedError('This method is abstract')

    def group_id(self) -> typing.Optional[str]:
        """ Return group id that unite records (in order 'ScheduledTaskPostponePolicy.keep_first' and
        'ScheduledTaskPostponePolicy.keep_last' to work; :meth:`.ScheduleRecordProto.simultaneous_policy` depends on
        this id also)

        :return: group id or None if this record is standalone
        """
        return None

    def ttl(self) -> typing.Union[int, float, None]:
        """ Return unix time when this record should be discarded

        :return: unix time in seconds or None if this record can not be expired. A timestamp must be return in UTC
        """
        return None

    def simultaneous_runs(self) -> int:
        """ Return how many records with the same group id may be run simultaneously. If non-positive value is return
        then there is no restrictions
        """
        return 0

    def postpone_policy(self) -> ScheduledTaskPostponePolicy:
        """ Return a postpone policy
        """
        return ScheduledTaskPostponePolicy.wait


@enum.unique
class SchedulerFeedback(enum.Enum):
    """ This enum flag is used along with the :meth:`.SchedulerProto.scheduler_feedback` method
    """
    source_subscribed = enum.auto()    # this flag shows that source in subscribed
    source_unsubscribed = enum.auto()  # this flag shows that source in unsubscribed


class ScheduleSourceProto(CapabilitiesAndSignals):
    """ This class may generate :class:`.ScheduleRecordProto` requests for a scheduler (:class:`.SchedulerProto`).
    This class decides what tasks and when should be run. When a time is come then this source emits
    a ScheduleSourceProto.task_scheduled signal
    """

    task_scheduled = Signal(ScheduleRecordProto)   # a new task should be started

    @capability
    def scheduler_feedback(self, scheduler: 'SchedulerProto', feedback: SchedulerFeedback) -> None:
        """ This method is used with the :class:`.SchedulerProto` class to notify that this source is used
        by the specified scheduler

        :param scheduler: scheduler that notifies
        :param feedback: scheduler's feedback event
        """
        raise NotImplementedError('The "stop" method is not supported')


# noinspection PyAbstractClass
class SchedulerProto(SignalSource):
    """ Represent a scheduler. A class that is able to execute tasks (:class:`.ScheduleRecordProto`) scheduled
    by sources (:class:`.ScheduleSourceProto`). This class tracks state of tasks that are running
    """

    task_scheduled = Signal(ScheduleRecordProto)            # a new task received from some source
    scheduled_task_dropped = Signal(ScheduleRecordProto)    # a scheduled task dropped and would not start
    scheduled_task_postponed = Signal(ScheduleRecordProto)  # a scheduled task dropped and will start later
    scheduled_task_expired = Signal(ScheduleRecordProto)    # a scheduled task dropped because of expired ttl
    scheduled_task_started = Signal(ScheduleRecordProto)    # a scheduled task started
    scheduled_task_completed = Signal(ScheduleRecordProto)  # a scheduled task completed

    @abstractmethod
    def subscribe(self, schedule_source: ScheduleSourceProto) -> None:
        """ Subscribe this scheduler to a specified source in order to start tasks from it

        :param schedule_source: source of records that should be subscribed
        """
        raise NotImplementedError('This method is abstract')

    @abstractmethod
    def unsubscribe(self, schedule_source: ScheduleSourceProto) -> None:
        """ Unsubscribe this scheduler from a specified sources and do process tasks from it

        :param schedule_source: source of records to unsubscribe from
        """
        raise NotImplementedError('This method is abstract')


class TaskExecutorProto(metaclass=ABCMeta):
    """ This is a class that executes tasks
    """

    @abstractmethod
    def submit_task(self, task: TaskProto) -> bool:
        """ Try to execute a task and return True if the task will be/or is already executed (return False otherwise)

        :param task: a task to execute
        """
        raise NotImplementedError('This method is abstract')

    @abstractmethod
    def complete_task(self, task: TaskProto) -> bool:
        """ Should be called for every successfully submitted task in order to finalize it

        :param task: task to finalize
        :return: True if a task is finalized successfully or return False otherwise
        """
        raise NotImplementedError('This method is abstract')

    @abstractmethod
    def wait_task(
        self,
        task: TaskProto,
        timeout: typing.Union[int, float, None] = None
    ) -> bool:
        """ Wait for a task completion

        :param task: a successfully submitted task to wait
        :param timeout: defines whether this function should be called in a blocking manner (and for how long)
        If value is None then this function will wait forever, if value is negative or zero, then this function will
        poll current state, otherwise -- number of seconds to wait for

        :return: return True if the task is completed and return False otherwise
        """
        raise NotImplementedError('This method is abstract')

    async def start_async(self, task: TaskProto) -> None:
        """Start a task and wait for a task to complete in async-way

        :param task: a task to start and wait
        """
        raise NotImplementedError('This method is abstract')

    @abstractmethod
    def tasks(self) -> typing.Generator[TaskProto, None, None]:
        """ Return generator that yields currently running tasks
        """
        raise NotImplementedError('This method is abstract')



# -*- coding: utf-8 -*-
# pyknic/lib/signals/extra.py
#
# Copyright (C) 2024 the pyknic authors and contributors
# <see AUTHORS file>
#
# This file is part of pyknic.
#
# pyknic is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# pyknic is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with pyknic.  If not, see <http://www.gnu.org/licenses/>.

import asyncio
import threading
import types
import typing
import weakref

from dataclasses import dataclass

from pyknic.lib.verify import verify_value

from pyknic.lib.signals.proto import SignalSourceProto, SignalCallbackType, Signal
from pyknic.lib.thread import CriticalResource


class BoundedCallback:
    """ This class allows to use a bounded method as a signal callback
    """

    @verify_value(callback=lambda x: hasattr(x, '__self__') and hasattr(x, '__func__'))
    def __init__(self, callback: SignalCallbackType):
        """ Create a bounded callback

        :param callback: a function to call
        """
        self.__callback_self = weakref.ref(
            callback.__self__  # type: ignore[attr-defined]  # callback may have __self__
        )
        self.__callback = callback.__func__  # type: ignore[attr-defined]  # callback may have __func__

    def __call__(self, source: SignalSourceProto, signal: Signal, value: typing.Any) -> None:
        """ Execute original function
        """
        s = self.__callback_self()
        if s:
            self.__callback(s, source, signal, value)


class CallbackWrapper:
    """ This is a base class for a simple wrapper factory. Is mostly used by :class:`.SignalProxyProto` implementations
    """

    def __init__(self, callback: SignalCallbackType, weak_callback: bool = False):
        """ Create a new wrapper

        :param callback: original callback to execute
        :param weak_callback: whether a general link or a weak link should be made for a callback
        """
        self.__weak = weak_callback
        self.__callback = weakref.ref(callback) if weak_callback else callback

    def _pre_hook(
        self, callback: SignalCallbackType, source: SignalSourceProto, signal: Signal, value: typing.Any
    ) -> bool:
        """ This method is called before a callback

        :param callback: a callback this wrapper has
        :param source: signal origin
        :param signal: emitted signal
        :param value: value that was emitted with a signal

        :return: True if callback and post hook should be called or False otherwise
        """
        return True

    def __call__(self, source: SignalSourceProto, signal: Signal, value: typing.Any) -> None:
        """ Execute an original callback

        :param source: signal origin
        :param signal: emitted signal
        :param value: value that was emitted with a signal
        """
        callback_obj: SignalCallbackType = (
            self.__callback() if self.__weak else self.__callback  # type: ignore[call-arg, assignment]  # mypy issue
        )
        if callback_obj is not None:
            if self._pre_hook(callback_obj, source, signal, value):
                callback_obj(source, signal, value)
                self._post_hook(callback_obj, source, signal, value)

    def _post_hook(
        self, callback: SignalCallbackType, source: SignalSourceProto, signal: Signal, value: typing.Any
    ) -> None:
        """ This method is called after a callback

        :param callback: a callback this wrapper has
        :param source: signal origin
        :param signal: emitted signal
        :param value: value that was emitted with a signal
        """
        pass

    @classmethod
    def wrapper(cls, callback: SignalCallbackType, weak_callback: bool = False) -> 'CallbackWrapper':
        """ Factory class method that returns class instance

        :param callback: same as the "callback" in the :meth:`.CallbackWrapper.__init__` method
        :param weak_callback: same as the "weak_callback" in the :meth:`.CallbackWrapper.__init__` method
        """
        return cls(callback, weak_callback=weak_callback)


class SignalResender:
    """ This class helps to resend signals with a new origin (and/or signal and/or value =) )
    """

    def __init__(
        self,
        target_source: SignalSourceProto,
        target_signal: typing.Optional[Signal] = None,
        weak_target: bool = False,
        value_converter: typing.Optional[typing.Callable[[SignalSourceProto, Signal, typing.Any], typing.Any]] = None
    ):
        """ Create a callback that re-emits signal with a new origin

        :param target_source: source that will send signal as a new origin
        :param target_signal: signal to send (if not defined then the emitted "signal" will be reused)
        :param weak_target: whether a target reference should be kept as a weak reference
        :param value_converter: this callback (if defined) will translate received signal to a signal's value
        that will be used
        """
        self.__weak_target = weak_target
        self.__target_source = weakref.ref(target_source) if weak_target else target_source
        self.__target_signal = target_signal
        self.__converter = value_converter if value_converter else self.__default_converter

    def __default_converter(self, source: SignalSourceProto, signal: Signal, value: typing.Any) -> typing.Any:
        """ Default converter that returns value as is
        """
        return value

    def __call__(self, source: SignalSourceProto, signal: Signal, value: typing.Any) -> None:
        """ A callback that to the job
        """
        target_src = self.__target_source
        target: SignalSourceProto = (
            target_src() if self.__weak_target else target_src  # type: ignore[operator, assignment]  # mypy issue
        )
        if target is not None:
            sig = self.__target_signal if self.__target_signal else signal
            target.emit(sig, self.__converter(source, sig, value))


class CallbacksHolder:
    """ This class helps to store callbacks with respect to memory
    """

    def __init__(self) -> None:
        """ Create a new holder to store callbacks
        """
        self.__callbacks: weakref.WeakValueDictionary[SignalCallbackType, typing.Any] = weakref.WeakValueDictionary()

    def keep_callback(self, callback: SignalCallbackType, reference_key: typing.Any) -> SignalCallbackType:
        """ Keep a new callback and return it

        :param callback: a callback to store
        :param reference_key: an object. While this object exists a callback exists
        """
        self.__callbacks[callback] = reference_key
        return callback


class CustomizedCallback:
    """ This class helps to call a callback with additional static arguments
    """

    class StaticWrapper:
        """ This is a wrapper for a callback
        """

        def __init__(
            self,
            callback: typing.Callable[[SignalSourceProto, Signal, typing.Any], None],
            **kwargs: typing.Any
        ) -> None:
            """ Create a wrapper

            :param callback: an original callback to call
            :param kwargs: static arguments to use along with the call
            """
            self.__callback = callback
            self.__kwargs = kwargs

        def __call__(self, source: SignalSourceProto, signal: Signal, value: typing.Any) -> None:
            """ Execute an original callback with extra static arguments
            """
            self.__callback(source, signal, value, **self.__kwargs)

    def __init__(self) -> None:
        """ Create a holder that will create callbacks with static arguments
        """
        self.__holder = CallbacksHolder()

    def customize(
        self,
        callback: typing.Callable[[SignalSourceProto, Signal, typing.Any], None],
        reference_key: typing.Any,
        **kwargs: typing.Any
    ) -> SignalCallbackType:
        """ Create and store a new callback

        :param callback: a base callback to use
        :param reference_key: a reference key to store a callback (the same as the "reference_key" parameter
        in the :meth:`.CallbacksHolder.keep_callback` method)
        :param kwargs: static arguments to use with the callback
        """
        custom_callback = CustomizedCallback.StaticWrapper(callback, **kwargs)
        self.__holder.keep_callback(custom_callback, reference_key)
        return custom_callback


@dataclass
class ReceivedSignal:
    """This dataclass represents a received signal."""
    source: SignalSourceProto  # signal origin
    signal: Signal             # emitted signal
    value: typing.Any          # value of a signal


class SignalWaiter:
    """ This class helps to wait for a signal
    """

    def __init__(
        self,
        target_source: SignalSourceProto,
        target_signal: Signal,
        timeout: typing.Union[int, float, None] = None,
        event: typing.Optional[threading.Event] = None,
        value_matcher: typing.Optional[typing.Callable[[typing.Any], bool]] = None
    ):
        """ Create object that will wait for a signal and sign up for it

        :param target_source: a source that will be subscribed for a signal to come
        :param target_signal: a signal to await
        :param timeout: timeout with which a result will be awaited. If the timeout is None then wait forever
        :param event: optional event that will be used to wait a signal. If not defined then internal event is used
        :param value_matcher: optional callable that checks received signal value. If defined then
        the event will be set only if this function return True on received signal value
        """
        self.__timeout = timeout
        self.__event = event if event else threading.Event()
        self.__matcher = value_matcher
        self.__received_signal: typing.Optional[ReceivedSignal] = None

        target_source.callback(target_signal, self)

    def __enter__(self) -> 'SignalWaiter':
        """ Start a context
        """
        return self

    def __exit__(
        self,
        exc_type: typing.Optional[typing.Type[BaseException]],
        exc_val: typing.Optional[BaseException],
        exc_tb: typing.Optional[types.TracebackType]
    ) -> None:
        """ Exit a context and wait for a signal
        """
        if exc_type is None:
            if self.wait() is None:
                raise TimeoutError("Signal wasn't received in time")

    def __call__(self, source: SignalSourceProto, signal: Signal, value: typing.Any) -> None:
        """ A callback that may be subscribed to more signals to come
        """
        if not self.__received_signal:
            if self.__matcher is None or self.__matcher(value):
                self.__received_signal = ReceivedSignal(source, signal, value)
                self.__event.set()

    def wait(self, ) -> typing.Optional[ReceivedSignal]:
        """ Wait for a signal and return True if signal has been received or False otherwise
        """
        self.__event.wait(self.__timeout)
        return self.__received_signal


class AsyncWatchDog(CriticalResource):
    """ This class helps to wait for a signal in asynchronous way
    """

    def __init__(
        self,
        loop: asyncio.AbstractEventLoop,
        target_source: SignalSourceProto,
        target_signal: Signal,
        cr_timeout: typing.Union[int, float, None] = None,
        value_matcher: typing.Optional[typing.Callable[[typing.Any], bool]] = None,
    ):
        """ Create object that will wait for a signal and sign up for it

        :param loop: event loop that will wait for a signal
        :param target_source: a source that will be subscribed for a signal to come
        :param target_signal: a signal to await
        :param cr_timeout: timeout to use as a critical resource synchronization period
        :param value_matcher: optional callable that checks received signal value. If defined then
        the event will be set only if this function return True on received signal value
        """
        CriticalResource.__init__(self, cr_timeout)

        self.__loop = loop
        self.__matcher = value_matcher

        self.__received_signal: typing.Optional[ReceivedSignal] = None
        self.__watchdog_future: typing.Optional[asyncio.Future[ReceivedSignal]] = None

        target_source.callback(target_signal, self)

    def __call__(self, source: SignalSourceProto, signal: Signal, value: typing.Any) -> None:
        """ A callback that may be subscribed to more signals to come
        """
        if not self.__received_signal:
            if self.__matcher is None or self.__matcher(value):

                with self.critical_context():
                    if self.__received_signal is None:
                        self.__received_signal = ReceivedSignal(source, signal, value)
                        if self.__watchdog_future is not None:
                            self.__loop.call_soon_threadsafe(
                                self.__watchdog_future.set_result, self.__received_signal
                            )

    async def watchdog_future(self) -> asyncio.Future[ReceivedSignal]:
        """Return future object that will wait for a signal to come."""
        current_loop = asyncio.get_event_loop()
        if current_loop is not self.__loop:
            raise RuntimeError('Watchdog has been created in another loop')

        with self.critical_context():
            if self.__watchdog_future is None:
                self.__watchdog_future = asyncio.Future(loop=self.__loop)
            if not self.__watchdog_future.done() and self.__received_signal is not None:
                self.__watchdog_future.set_result(self.__received_signal)

        return self.__watchdog_future

    async def wait(
       self, timeout: typing.Optional[typing.Union[int, float]] = None
    ) -> asyncio.Future[ReceivedSignal]:
        """Wait for a signal and return a future-object. This object should be checked whether
        a signal has been received or timeout has been passed

        :param timeout: timeout with which a result will be awaited. If the timeout is None then wait forever
        """
        watchdog = await self.watchdog_future()
        asyncio_wait_tasks: typing.List[
            typing.Union[
                asyncio.Future[ReceivedSignal],
                asyncio.Task[None]
            ]
        ] = [watchdog]

        timeout_task = None
        if timeout is not None:
            timeout_task = asyncio.create_task(asyncio.sleep(timeout))
            asyncio_wait_tasks.append(timeout_task)

        done, pending = await asyncio.wait(asyncio_wait_tasks, return_when=asyncio.FIRST_COMPLETED)

        if timeout_task is not None and timeout_task in pending:
            timeout_task.cancel()

        return watchdog



# -*- coding: utf-8 -*-
# pyknic/lib/tasks/plain_task.py
#
# Copyright (C) 2024 the pyknic authors and contributors
# <see AUTHORS file>
#
# This file is part of pyknic.
#
# pyknic is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# pyknic is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with pyknic.  If not, see <http://www.gnu.org/licenses/>.

import typing

from pyknic.lib.tasks.proto import TaskProto, TaskResult


class PlainTask(TaskProto):
    """ This class may be treated as an adapter from any callable object to the `.TaskProto` class
    """

    def __init__(self, fn: typing.Callable[[], typing.Any]):
        """ Create a new task that executed the specified object

        :param fn: object to execute
        """
        TaskProto.__init__(self)
        self.__function = fn

    def start(self) -> None:
        """ The :meth:`.TaskProto.start` method implementation
        """
        self.emit(self.task_started)
        try:
            result = self.__function()
        except Exception as e:
            self.emit(self.task_completed, TaskResult(None, exception=e))
            return

        self.emit(self.task_completed, TaskResult(result))



# -*- coding: utf-8 -*-

import typing

from pyknic.lib.tasks.proto import ScheduleRecordProto, ScheduledTaskPostponePolicy
from pyknic.lib.tasks.scheduler.record import ScheduleRecord

if typing.TYPE_CHECKING:
    # noinspection PyUnresolvedReferences
    from conftest import SampleTasks


class TestScheduleRecord:

    def test(self, sample_tasks: 'SampleTasks') -> None:
        task1 = sample_tasks.DummyTask()
        source1 = sample_tasks.DummySource()
        record = ScheduleRecord(task1, source1)
        assert(isinstance(record, ScheduleRecord) is True)
        assert(isinstance(record, ScheduleRecordProto) is True)
        assert(record.task() is task1)
        assert(record.source() is source1)
        assert(record.group_id() is None)
        assert(record.ttl() is None)
        assert(record.simultaneous_runs() == 0)
        assert(record.postpone_policy() == ScheduledTaskPostponePolicy.wait)

        task2 = sample_tasks.DummyTask()
        source2 = sample_tasks.DummySource()
        record = ScheduleRecord(
            task2,
            source2,
            group_id='task_group',
            ttl=10,
            simultaneous_runs=2,
            postpone_policy=ScheduledTaskPostponePolicy.drop
        )

        assert(record.task() is task2)
        assert(record.source() is source2)
        assert(record.group_id() == 'task_group')
        assert(record.ttl() == 10)
        assert(record.simultaneous_runs() == 2)
        assert(record.postpone_policy() == ScheduledTaskPostponePolicy.drop)



# -*- coding: utf-8 -*-
# pyknic/tasks/config.py
#
# Copyright (C) 2024 the pyknic authors and contributors
# <see AUTHORS file>
#
# This file is part of pyknic.
#
# pyknic is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# pyknic is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with pyknic.  If not, see <http://www.gnu.org/licenses/>.

# TODO: write tests for the code

import itertools
import os
import typing

from pyknic.lib.tasks.scheduler.chain_source import ChainedTask, __default_chained_tasks_registry__
from pyknic.lib.registry import register_api

from pyknic.lib.log import Logger
from pyknic.lib.config import Config


@register_api(__default_chained_tasks_registry__, "config_task")
class ConfigTask(ChainedTask):
    """ Generate a configuration

    Configuration is generated by applying multiple files. Files are applied in the following order:
      - default configurations (config.yaml files from the "pyknic.tasks" subdirectories)
      - if the "PYKNIC_APP_DIR_CONFIG" variable is set then read configurations (any file) from this directory
        and its subdirectories
      - if the "PYKNIC_APP_FILE_CONFIG" variable is set then read configuration from this file
    """

    __apps_defaults_dir__ = os.path.join(os.path.dirname(__file__), '..')  # directory with default configurations
    __default_file__ = "config.yaml"  # default configuration file name

    __app_dir_config_envar__ = "PYKNIC_APP_DIR_CONFIG"  # defines an environment variable for directory with configs
    __app_file_config_envar__ = "PYKNIC_APP_FILE_CONFIG"  # defines an environment variable for config file

    def __load_file(self, config: Config, filename: str) -> None:
        """ Merge data from a file to a config

        :param config: a config to merge to
        :param filename: a file that holds data
        """
        if os.path.isfile(filename) is False:
            raise ValueError(f'Invalid configuration (no such file): "{filename}"')

        with open(filename) as f:
            config.merge_file(f)
        Logger.info(f'Configuration loaded from the file: "{os.path.abspath(filename)}"')

    def __check_directory(
        self, config: Config, dir_path: str, file_pattern: typing.Optional[str] = None
    ) -> typing.Set[str]:
        """ List directory and merge found files with a config

        :param config: a config to merge to
        :param dir_path: a directory to search in
        :param file_pattern: if defined only this file will be read from directories

        :return: return absolute paths for found inner directories
        """
        inner_dirs = set()

        for e in os.listdir(dir_path):
            e_path = os.path.join(dir_path, e)
            Logger.debug(f"Checking configuration in the directory \"{os.path.relpath(dir_path)}\" -- {e_path}")

            if os.path.isdir(e_path):
                inner_dirs.add(e_path)
            elif file_pattern is not None and e == file_pattern:
                self.__load_file(config, e_path)
            elif file_pattern is None and os.path.isfile(e_path):
                self.__load_file(config, e_path)

        return inner_dirs

    def __read_directory(self, config: Config, dir_path: str, file_pattern: typing.Optional[str] = None) -> None:
        """ Recursively list directories and merge found files with a config

        :param config: a config to merge to
        :param dir_path: a root directory inside which files will be searched
        :param file_pattern: if defined only this file will be read from directories
        """
        check_dirs = [dir_path]
        while check_dirs:
            check_dirs.sort()
            next_dirs = [self.__check_directory(config, x, file_pattern=file_pattern) for x in check_dirs]
            check_dirs = list(itertools.chain(*next_dirs))  # flatten the list

    def start(self) -> None:
        """ The :meth:`.TaskProto.start` method implementation
        """
        self.wait_for('log_task')
        Logger.info('Reading configuration')

        result = Config()
        self.__read_directory(result, self.__apps_defaults_dir__, self.__default_file__)

        if self.__app_dir_config_envar__ in os.environ:
            self.__read_directory(result, os.environ[self.__app_dir_config_envar__])

        if self.__app_file_config_envar__ in os.environ:
            self.__load_file(result, os.environ[self.__app_file_config_envar__])

        self.save_result(result)
        Logger.info('Configuration loaded')

    def task_name(self) -> typing.Optional[str]:
        """ The :meth:`.ChainedTask.task_name` method implementation
        """
        return 'config-reader'

    @classmethod
    def dependencies(cls) -> typing.Optional[typing.Set[str]]:
        """ The :meth:`.ChainedTask.dependencies` method implementation
        """
        return {"log_task"}



# -*- coding: utf-8 -*-
# pyknic/lib/backup/archive_v1.py
#
# Copyright (C) 2025 the pyknic authors and contributors
# <see AUTHORS file>
#
# This file is part of pyknic.
#
# pyknic is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# pyknic is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with pyknic.  If not, see <http://www.gnu.org/licenses/>.

# TODO: Add snapshot support and keep original lv uuid to metadata

import asyncio
import base64
import copy
import functools
import enum
import io
import math
import time
import typing

import pydantic

from pyknic.lib.io import IOGenerator
from pyknic.lib.io.aio_wrapper import IOThrottler, chain_sync_processor, cg
from pyknic.lib.crypto.hash import __default_hashers_registry__
from pyknic.lib.io.compression import __default_io_compressors_registry__
from pyknic.lib.crypto.cipher import __default_cipher_registry__
from pyknic.lib.crypto.kdf import PBKDF2
from pyknic.lib.crypto.padding import PKCS7Padding
from pyknic.lib.io.stats import GeneratorStats
from pyknic.lib.io.tar import TarArchive, TarInnerDynamicGenerator, TarInnerFileGenerator
from pyknic.lib.verify import verify_value


@enum.unique
class ArchiveInnerFiles(enum.Enum):
    """This enum describes possible file names inside an archive.
    """
    header_meta = 'header_meta.json'  # name of file that has the :class:`.ArchiveV1HeaderMeta`
    tail_meta = 'tail_meta.json'      # name of file that has the :class:`.ArchiveV1TailMeta`

    @classmethod
    def backup_file(cls, header_meta: 'ArchiveV1HeaderMeta') -> str:
        """Return name of a file for archiving data
        """

        file_name = 'backup'

        match header_meta.type:
            case ArchiveType.io_archive:
                pass
            case ArchiveType.file_archive:
                file_name += '.tar'
            case _:
                raise ValueError(f'Unknown archive type: {header_meta.type}')

        match header_meta.compression:
            case CompressionMode.no_compression:
                pass
            case CompressionMode.gzip:
                file_name += '.gz'
            case CompressionMode.bzip2:
                file_name += '.bz2'
            case CompressionMode.lzma:
                file_name += '.xz'
            case _:
                raise ValueError(f'Unknown compression type: {header_meta.compression}')

        if header_meta.cipher:
            file_name += '.enc'

        return file_name


@enum.unique
class ArchiveType(enum.Enum):
    """This enum describes possible archive types.
    """
    io_archive = 'io_archive'      # archive has inner file that stores data as is
    file_archive = 'file_archive'  # archive has inner tar archive that consists of files from FS


@enum.unique
class CompressionMode(enum.Enum):
    """This enum describes possible compression modes. Values are pretty obvious.
    """
    no_compression = 'no_compression'
    gzip = 'gzip'
    bzip2 = 'bzip2'
    lzma = 'lzma'


@enum.unique
class HashMethod(enum.Enum):
    """This enum describes possible hashing methods. Values are pretty obvious.
    """
    blake2b_64 = 'blake2b_64'
    blake2s_32 = 'blake2s_32'
    md5        = 'md5'         # noqa: E221
    sha1       = 'sha1'        # noqa: E221
    sha224     = 'sha224'      # noqa: E221
    sha256     = 'sha256'      # noqa: E221
    sha384     = 'sha384'      # noqa: E221
    sha512     = 'sha512'      # noqa: E221
    sha512_224 = 'sha512_224'  # noqa: E221
    sha512_256 = 'sha512_256'  # noqa: E221
    sha3_224   = 'sha3_224'    # noqa: E221
    sha3_256   = 'sha3_256'    # noqa: E221
    sha3_384   = 'sha3_384'    # noqa: E221
    sha3_512   = 'sha3_512'    # noqa: E221


class ArchiveV1PBKDF(pydantic.BaseModel):
    """This model describes the way a secret key is generated.
    """
    salt: pydantic.Base64Bytes
    iterations: int
    hash_name: str


class ArchiveV1MetaCipher(pydantic.BaseModel):
    """This model describes encryption details
    """
    pbkdf: ArchiveV1PBKDF
    cipher_name: str
    decryptor_info: typing.Any


class ArchiveV1HeaderMeta(pydantic.BaseModel):
    """This model describes main archive information that may be collected before a start
    """
    version: typing.Literal['1.0.0'] = pydantic.Field(default='1.0.0', frozen=True)  # static version of this model
    type: ArchiveType                                                                # archive type -- io or file mode
    created: int = pydantic.Field(default_factory=lambda _: int(time.time()))        # when archive was created
    # as a timestamp
    compression: CompressionMode                                                     # compression details
    cipher: typing.Optional[ArchiveV1MetaCipher] = None                              # encryption details
    extra: typing.Optional[typing.Dict[str, typing.Any]] = None                      # any additional info


class ArchiveV1HashInfo(pydantic.BaseModel):
    """This model describes digest information about archived data
    """
    algorithm: HashMethod         # used algorithm
    digest: pydantic.Base64Bytes  # calculated digest


class ArchiveV1TailMeta(pydantic.BaseModel):
    """This model describes archived data
    """
    write_rate: int                                  # average data rate (bytes per second)
    hashes: typing.List[ArchiveV1HashInfo] = list()  # digests
    duration: int                                    # how long archiving process took (in seconds)


class _BackupHelper:
    """This class implements some basic logic for archiving implementation
    """

    __default_cipher__ = 'AES-256-CBC'  # default cipher that is used when encryption key is specified

    @verify_value(cipher_name=lambda x: x is None or __default_cipher_registry__.has(x))
    def __init__(
        self,
        archive_type: ArchiveType,
        encryption_key: typing.Optional[str] = None,
        cipher_name: typing.Optional[str] = None,
        compression: typing.Optional[CompressionMode] = None,
        hash_algorithms: typing.Optional[typing.Iterable[HashMethod]] = None,
        extra_meta: typing.Optional[typing.Dict[str, typing.Any]] = None
    ) -> None:
        """Create a helper

        :param archive_type: defines an archive type (just to be stored in archive meta)
        :param encryption_key: if specified then this is a base key that is used for data encryption. There will
        be no encryption of data by default.
        :param cipher_name: a cipher to use. The _BackupHelper.__default_cipher__ is used by default. For enabling
        encryption be sure that the encryption_key was set.
        :param compression: defines how archiving data will be compressed (meta-data will be always uncompressed).
        No compression by default
        :param hash_algorithms: algorithms to use for digests creation. May be used for future consistency checks.
        No digest will be provided by default.
        :param extra_meta: any additional info to store within archive meta-data
        """
        if extra_meta is not None:
            extra_meta = copy.deepcopy(extra_meta)

        self.__archive_type = archive_type
        self.__compression = compression if compression else CompressionMode.no_compression
        self.__hash_algorithms = hash_algorithms if hash_algorithms else []
        self.__extra_meta = extra_meta

        self.__cipher_data = None
        self.__cipher_encryptor = None
        if encryption_key is not None:
            if cipher_name is None:
                cipher_name = self.__default_cipher__

            cipher_cls = __default_cipher_registry__.get(cipher_name)

            pbkdf = PBKDF2(encryption_key.encode(), derived_key_length=cipher_cls.key_size())

            self.__cipher_encryptor = cipher_cls.create_encryptor(pbkdf.derived_key())
            self.__cipher_data = ArchiveV1MetaCipher(
                pbkdf=ArchiveV1PBKDF(
                    salt=base64.b64encode(pbkdf.salt()),
                    iterations=pbkdf.iterations(),
                    hash_name=pbkdf.hash_name()
                ),
                cipher_name=cipher_name,
                decryptor_info=self.__cipher_encryptor.decryptor_init_data(),
            )

        self.__digests = dict()
        for algorithm in self.__hash_algorithms:
            self.__digests[algorithm] = __default_hashers_registry__.get(algorithm.value)()

        self.__write_stats = GeneratorStats()
        self.__start_time = time.monotonic()

    def header(self) -> ArchiveV1HeaderMeta:
        """Create and return main meta-data
        """
        return ArchiveV1HeaderMeta(
            type=self.__archive_type,
            compression=self.__compression,
            cipher=self.__cipher_data,
            extra=self.__extra_meta
        )

    def tail_data(self) -> IOGenerator:
        """Create and return meta-data with some statistics and digests (if any)
        """
        hashes = [
            ArchiveV1HashInfo(algorithm=h, digest=base64.b64encode(d.digest())) for h, d in self.__digests.items()
        ]

        tail_metadata = ArchiveV1TailMeta(
            write_rate=math.ceil(self.__write_stats.rate()),
            hashes=hashes,
            duration=math.ceil(time.monotonic() - self.__start_time)
        )

        yield tail_metadata.model_dump_json().encode()

    def backup_chain(self, data: IOGenerator) -> IOGenerator:
        """Process archiving data and return encrypted and/or compressed and/or raw result (depends on arguments
        with which object was created)

        :param data: data to backup
        """
        backup_processors = []

        if self.__compression and self.__compression != CompressionMode.no_compression:
            backup_processors += [
                __default_io_compressors_registry__.get(self.__compression.value)().compress
            ]

        if self.__cipher_encryptor:
            padding_obj = PKCS7Padding()
            backup_processors += [
                functools.partial(padding_obj.pad, block_size=self.__cipher_encryptor.algo_block_size()),
                self.__cipher_encryptor.encrypt
            ]

        backup_processors += [x.update for x in self.__digests.values()]
        backup_processors += [self.__write_stats.process]

        return chain_sync_processor(data, *backup_processors)

    @classmethod
    def hashes_validate_chain(cls, tail: ArchiveV1TailMeta, data: IOGenerator) -> IOGenerator:
        """ Check that data inside archive has the correct hashes

        :param tail: meta-data with digests
        :param data: data to validate
        """

        digests = dict()
        for i in tail.hashes:
            digests[i.algorithm] = __default_hashers_registry__.get(i.algorithm.value)()

        digests_processors = [x.update for x in digests.values()]
        if len(digests_processors) == 0:
            raise ValueError('There is no digests inside archive!')

        yield from chain_sync_processor(data, *digests_processors)

        for i in tail.hashes:
            calculated_hash = digests[i.algorithm]

            if calculated_hash.digest() != i.digest:
                raise ValueError(f'The calculated hash for the "{i.algorithm.value}" did not match!')

    @classmethod
    def unarchive_chain(
        cls,
        header: ArchiveV1HeaderMeta,
        tail: ArchiveV1TailMeta,
        data: IOGenerator,
        encryption_key: typing.Optional[str] = None,
        validate_hashes: bool = False,
    ) -> IOGenerator:
        """Process archived data and return decrypted and/or uncompressed and/or raw result (depends on arguments
        with which archive was created)

        :param header: original archive header meta-data
        :param tail: original archive tail meta-data
        :param data: data to restore
        :param encryption_key: encryption key that will be used for decryption
        :param validate_hashes: whether to validate the archive hashes

        :note: there is no digest checking at the moment!
        """

        backup_processors = []

        if validate_hashes:
            backup_processors += [functools.partial(cls.hashes_validate_chain, tail)]

        if header.cipher is not None:
            if encryption_key is None:
                raise ValueError('Archive is encrypted, but no encryption key provided')

            padding_obj = PKCS7Padding()
            cipher_cls = __default_cipher_registry__.get(header.cipher.cipher_name)

            pbkdf = PBKDF2(
                encryption_key.encode(),
                derived_key_length=cipher_cls.key_size(),
                salt=header.cipher.pbkdf.salt,
                iterations_count=header.cipher.pbkdf.iterations,
                hash_fn_name=header.cipher.pbkdf.hash_name
            )

            cipher_decryptor = cipher_cls.create_decryptor(pbkdf.derived_key(), header.cipher.decryptor_info)

            backup_processors += [
                cipher_decryptor.decrypt,
                functools.partial(padding_obj.undo_pad, block_size=cipher_cls.algo_block_size()),
            ]

        if header.compression and header.compression != CompressionMode.no_compression:
            backup_processors += [
                __default_io_compressors_registry__.get(header.compression.value)().decompress
            ]

        return chain_sync_processor(data, *backup_processors)


class BackupArchiveV1:
    """This class implements public methods for backing up and restore archives
    """

    def __init__(
        self,
        hash_algorithms: typing.Optional[typing.Iterable[HashMethod]] = None,
        extra_meta: typing.Optional[typing.Dict[str, typing.Any]] = None,
        throttling: typing.Optional[int] = None,
        compression: typing.Optional[CompressionMode] = None,
        encryption_key: typing.Optional[str] = None,
        cipher_name: typing.Optional[str] = None
    ):
        """Create backup archiver

        :param hash_algorithms: algorithms to use for digests creation. May be used for future consistency checks.
        No digest will be provided by default.
        :param extra_meta: any additional info to store within archive meta-data
        :param throttling: target data-rate (bytes per second) for saving archive
        :param compression: defines how archiving data will be compressed (meta-data will be always uncompressed).
        No compression by default
        :param encryption_key: if specified then this is a base key that is used for data encryption. There will
        be no encryption of data by default.
        :param cipher_name: a cipher to use. The _BackupHelper.__default_cipher__ is used by default. For enabling
        encryption be sure that the encryption_key was set.
        """
        self.__hashes = hash_algorithms
        self.__extra_meta = extra_meta
        self.__throttling = throttling
        self.__compression = compression
        self.__encryption_key = encryption_key
        self.__cipher_name = cipher_name

    async def __backup(
        self, archive_type: ArchiveType, data_reader: IOGenerator, destination: typing.IO[bytes]
    ) -> None:
        """This method wraps backup routine

        :param archive_type: type of archive to create
        :param data_reader: data to backup
        :param destination: file target archive to write to
        """
        helper = _BackupHelper(
            archive_type=archive_type,
            hash_algorithms=self.__hashes,
            extra_meta=self.__extra_meta,
            compression=self.__compression,
            encryption_key=self.__encryption_key,
            cipher_name=self.__cipher_name
        )

        header = helper.header()
        header_reader = IOThrottler.sync_reader(io.BytesIO(header.model_dump_json().encode()))

        sources = [
            TarInnerDynamicGenerator(header_reader, ArchiveInnerFiles.header_meta.value),
            TarInnerDynamicGenerator(helper.backup_chain(data_reader), ArchiveInnerFiles.backup_file(header)),
            TarInnerDynamicGenerator(helper.tail_data(), ArchiveInnerFiles.tail_meta.value)
        ]

        await TarArchive().dynamic_archive(destination, sources, write_throttling=self.__throttling)

    async def backup_io(self, source: IOGenerator, destination: typing.IO[bytes]) -> None:
        """Backup dynamic data

        :param source: data to backup
        :param destination: file target archive to write to
        """
        await self.__backup(ArchiveType.io_archive, source, destination)

    async def backup_files(self, files: typing.Iterable[str], destination: typing.IO[bytes]) -> None:
        """Backup ordinary files

        :param files: files, directories, named sockets from FS to backup
        :param destination: file target archive to write to
        """
        # TODO: add note that check that there are no duplicates inside files generator

        def files_generator() -> typing.Generator[TarInnerFileGenerator, None, None]:
            for f in files:
                yield TarInnerFileGenerator(f)

        data_reader = TarArchive().static_archive(files_generator())

        await self.__backup(ArchiveType.file_archive, data_reader, destination)

    @classmethod
    def extract_header_meta(cls, archive: typing.IO[bytes]) -> ArchiveV1HeaderMeta:
        """Extract main meta-data from stored archive

        :param archive: archive to extract meta-data from
        """
        extractor = TarArchive().extract(archive, ArchiveInnerFiles.header_meta.value)
        header_buffer = io.BytesIO()
        cg(IOThrottler.sync_writer(extractor, header_buffer))

        return ArchiveV1HeaderMeta.model_validate_json(header_buffer.getvalue())

    @classmethod
    def extract_tail_meta(cls, archive: typing.IO[bytes]) -> ArchiveV1TailMeta:
        """Extract additional meta-data from stored archive

        :param archive: archive to extract meta-data from
        """
        extractor = TarArchive().extract(archive, ArchiveInnerFiles.tail_meta.value)
        header_buffer = io.BytesIO()
        cg(IOThrottler.sync_writer(extractor, header_buffer))

        return ArchiveV1TailMeta.model_validate_json(header_buffer.getvalue())

    @classmethod
    def extract_data(
        cls,
        archive: typing.IO[bytes],
        encryption_key: typing.Optional[str] = None,
        validate_hashes: bool = False,
    ) -> IOGenerator:
        """Extract data from stored archive

        :param archive: archive to extract from
        :param encryption_key: encryption key that was used for archive creation (if any)
        :param validate_hashes: whether to check hashes or not
        """
        header_meta = cls.extract_header_meta(archive)
        tail_meta = cls.extract_tail_meta(archive)
        inner_data = TarArchive().extract(archive, ArchiveInnerFiles.backup_file(header_meta))

        return _BackupHelper.unarchive_chain(
            header_meta, tail_meta, inner_data, encryption_key=encryption_key, validate_hashes=validate_hashes
        )

    @classmethod
    async def validate_archive(
        cls,
        archive: typing.IO[bytes],
        throttling: typing.Optional[int] = None
    ) -> None:
        """Validate data stored in an archive

        :param archive: archive to check
        :param throttling: bytes per second rate to read an archive
        """
        header_meta = cls.extract_header_meta(archive)
        tail_meta = cls.extract_tail_meta(archive)
        inner_data = TarArchive().extract(archive, ArchiveInnerFiles.backup_file(header_meta))

        throttler = IOThrottler(throttling=throttling)
        throttler.start()
        for chunk in _BackupHelper.hashes_validate_chain(tail_meta, inner_data):
            throttler += len(chunk)
            await asyncio.sleep(throttler.pause())



# -*- coding: utf-8 -*-
# pyknic/lib/verify.py
#
# Copyright (C) 2024 the pyknic authors and contributors
# <see AUTHORS file>
#
# This file is part of pyknic.
#
# pyknic is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# pyknic is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with pyknic.  If not, see <http://www.gnu.org/licenses/>.

import typing
import sys

from decorator import decorator
from inspect import getfullargspec, isclass, isfunction, getsource

from pyknic.lib.typing import GenericFunc, P, R


class Verifier:
    """ Base class for verifier implementation.

    Verifiers are classes, that generates decorators (which are later used for runtime function arguments
    checking). Derived classes (such as :class:`.TypeVerifier`, :class:`SubclassVerifier` and
    :class:`ValueVerifier`) check arguments for type and/or value validity. But each derived class uses
    its own syntax for check declaration (see :meth:`.Verifier.check`).

    Same checks can be grouped into one sentence if they are used for different arguments

    Checks can be simple (implemented by lambda-statements) or complex
    (implemented by standalone functions or classes). Because target function is decorated for checking it is
    possible to run checks sequentially.

    Example: ::

        @verify_type(a=int, b=str, d=(int, None), e=float)
        @verify_subclass(c=A, args=B)
        @verify_value(a=(lambda x: x > 5, lambda x: x < 10), args=lambda x: x > 0)
        @verify_value(c=lambda x: x.a == 'foo', d=lambda x: x is None or x < 0)
        def foo(a, b, c, d=None, *args, **kwargs):
            pass
    """

    def __init__(self, silent_checks: bool = False):
        """ Construct a new :class:`.Verifier`

        :param silent_checks: If it is not True, then debug information will be printed to stderr.
        """
        self._silent_checks = silent_checks

    def check(
        self,
        arg_spec: typing.Any,
        arg_name: str,
        decorated_function: typing.Callable[..., typing.Any]
    ) -> typing.Callable[..., typing.Any]:
        """ Return callable object that takes single value - future argument. This callable must raise
        an exception if error occurs. It is recommended to return None if everything is OK

        :param arg_spec: specification that is used to describe check like types, lambda-functions, \
        list of types just anything (see :meth:`.Verifier.decorator`)
        :param arg_name: parameter name from decorated function
        :param decorated_function: target function (function to decorate)
        """
        return lambda x: None

    def _args_checks_gen(
        self,
        decorated_function: typing.Callable[..., typing.Any],
        function_spec: typing.Any,
        arg_specs: typing.Any
    ) -> typing.Dict[str, typing.Callable[[typing.Any], None]]:
        """ Generate checks for positional argument testing

        :param decorated_function: function decorator
        :param function_spec: function inspect information
        :param arg_specs: argument specification (same as arg_specs in :meth:`.Verifier.decorate`)
        """
        inspected_args = function_spec.args
        args_check = {}

        for i in range(len(inspected_args)):
            arg_name = inspected_args[i]

            if arg_name in arg_specs.keys():
                args_check[arg_name] = self.check(arg_specs[arg_name], arg_name, decorated_function)
        return args_check

    def _args_checks_test(
        self,
        original_function: typing.Callable[..., typing.Any],
        function_spec: typing.Any,
        checks: typing.Dict[str, typing.Callable[[typing.Any], None]],
        args: typing.Tuple[typing.Any],
        arg_specs: typing.Any
    ) -> None:
        """ Test positional arguments by a generated checks

        :param original_function: original function, that was decorated
        :param function_spec: function inspect information
        :param checks: tests, that was generated by the :meth:`.Verifier._args_checks_gen`
        :param args: positional arguments to check (with varargs)
        :param arg_specs: argument specification (same as arg_specs in :meth:`.Verifier.decorate`)
        """
        inspected_args = function_spec.args

        for i in range(len(inspected_args)):
            param_name = inspected_args[i]
            if param_name in checks:
                try:
                    checks[param_name](args[i])
                except Exception as e:
                    self.help_info(e, original_function, param_name, arg_specs[param_name])
                    raise

    def _varargs_checks_gen(
        self,
        decorated_function: typing.Callable[..., typing.Any],
        function_spec: typing.Any,
        arg_specs: typing.Any
    ) -> typing.Optional[typing.Callable[..., typing.Any]]:
        """ Generate checks for positional variable argument (varargs) testing

        :param decorated_function: function decorator
        :param function_spec: function inspect information
        :param arg_specs: argument specification (same as arg_specs in :meth:`.Verifier.decorate`)
        """
        inspected_varargs = function_spec.varargs

        if inspected_varargs is not None and inspected_varargs in arg_specs.keys():
            return self.check(
                arg_specs[inspected_varargs], inspected_varargs, decorated_function
            )

        return None

    def _varargs_checks_test(
        self,
        original_function: typing.Callable[..., typing.Any],
        function_spec: typing.Any,
        check: typing.Optional[typing.Callable[..., typing.Any]],
        args: typing.Tuple[typing.Any],
        arg_specs: typing.Any
    ) -> None:
        """ Test varargs by a generated check

        :param original_function: original function, that was decorated
        :param function_spec: function inspect information
        :param check: test, that was generated by the :meth:`.Verifier._varargs_checks_gen`
        :param args: positional arguments to check (all of them)
        :param arg_specs: argument specification (same as arg_specs in :meth:`.Verifier.decorate`)
        """
        inspected_args = function_spec.args
        inspected_varargs = function_spec.varargs

        if inspected_varargs is not None and check is not None:
            for i in range(len(inspected_args), len(args)):
                try:
                    check(args[i])
                except Exception as e:
                    self.help_info(e, original_function, inspected_varargs, arg_specs[inspected_varargs])
                    raise

    def _kwargs_checks_gen(
        self,
        decorated_function: typing.Callable[..., typing.Any],
        function_spec: typing.Any,
        arg_specs: typing.Any
    ) -> typing.Dict[str, typing.Callable[[typing.Any], None]]:
        """ Generate checks for keyword argument testing

        :param decorated_function: function decorator
        :param function_spec: function inspect information
        :param arg_specs: argument specification (same as arg_specs in :meth:`.Verifier.decorate`)
        """
        args_names = []
        args_names.extend(function_spec.args)
        if function_spec.varargs is not None:
            args_names.append(function_spec.args)

        args_check = {}

        for arg_name in arg_specs.keys():
            if arg_name not in args_names:
                args_check[arg_name] = self.check(
                    arg_specs[arg_name], arg_name, decorated_function
                )
        return args_check

    def _kwargs_checks_test(
        self,
        original_function: typing.Callable[..., typing.Any],
        checks: typing.Dict[str, typing.Callable[[typing.Any], None]],
        kwargs: typing.Dict[str, typing.Any],
        arg_specs: typing.Any
    ) -> None:
        """ Test keyword arguments by a generated checks

        :param original_function: original function, that was decorated
        :param checks: tests, that was generated by the :meth:`.Verifier._kwargs_checks_gen`
        :param kwargs: keyword arguments to check
        :param arg_specs: argument specification (same as arg_specs in :meth:`.Verifier.decorate`)
        """

        for kw_key, kw_value in kwargs.items():
            if kw_key in checks.keys():
                try:
                    checks[kw_key](kw_value)
                except Exception as e:
                    self.help_info(e, original_function, kw_key, arg_specs[kw_key])
                    raise

    def decorator(
        self, **arg_specs: typing.Any
    ) -> typing.Callable[[GenericFunc[P, R]], GenericFunc[P, R]]:
        """ Return decorator that can decorate target function

        :param arg_specs: dictionary where keys are parameters name and values are theirs specification.\
        Specific specification is passed as is to :meth:`Verifier.check` method with corresponding \
        parameter name.
        """

        def first_level_decorator(
            decorated_function: typing.Callable[..., typing.Any]
        ) -> typing.Callable[..., typing.Any]:

            function_spec = getfullargspec(decorated_function)
            args_checks = self._args_checks_gen(decorated_function, function_spec, arg_specs)
            varargs_check = self._varargs_checks_gen(decorated_function, function_spec, arg_specs)
            kwargs_checks = self._kwargs_checks_gen(decorated_function, function_spec, arg_specs)

            def second_level_decorator(
                original_function: typing.Callable[..., typing.Any], *args: typing.Any, **kwargs: typing.Any
            ) -> typing.Any:
                self._args_checks_test(original_function, function_spec, args_checks, args, arg_specs)
                self._varargs_checks_test(original_function, function_spec, varargs_check, args, arg_specs)
                self._kwargs_checks_test(original_function, kwargs_checks, kwargs, arg_specs)

                return original_function(*args, **kwargs)
            return decorator(second_level_decorator)(decorated_function)  # type: ignore[no-any-return, no-untyped-call]
        return first_level_decorator

    def help_info(
        self,
        exc: BaseException,
        decorated_function: typing.Callable[..., typing.Any],
        arg_name: str,
        arg_spec: typing.Any
    ) -> None:
        """ Print debug information to stderr. (Do nothing if object was constructed with silent_checks=True)

        :param exc: raised exception
        :param decorated_function: target function (function to decorate)
        :param arg_name: function parameter name
        :param arg_spec: function parameter specification
        """
        if self._silent_checks is not True:
            print('Exception raised:', file=sys.stderr)
            print(str(exc), file=sys.stderr)
            fn_name = Verifier.function_name(decorated_function)
            print('Decorated function: %s' % fn_name, file=sys.stderr)
            if decorated_function.__doc__ is not None:
                print('Decorated function docstrings:', file=sys.stderr)
                print(decorated_function.__doc__, file=sys.stderr)
            print('Argument "%s" specification:' % arg_name, file=sys.stderr)
            if isfunction(arg_spec):
                print(getsource(arg_spec), file=sys.stderr)
            else:
                print(str(arg_spec), file=sys.stderr)
            print('', file=sys.stderr)

    @staticmethod
    def function_name(fn: typing.Callable[..., typing.Any]) -> str:
        """ Return function name in pretty style

        :param fn: source function
        :return: str
        """
        if hasattr(fn, '__qualname__'):
            return fn.__qualname__
        elif hasattr(fn, '__name__'):
            fn_name = fn.__name__
            if hasattr(fn, '__self__'):
                owner = fn.__self__
                if isclass(owner) is False:
                    owner = owner.__class__
                if hasattr(owner, '__name__'):
                    return f'{owner.__name__}.{fn_name}'
            return fn_name

        return f'Unknown function with id {id(fn)}'


class TypeVerifier(Verifier):
    """ Verifier that is used for type verification. Checks parameter if it is instance of specified class or
    classes. Specification accepts type or list/tuple/set of types

    Example:

        @verify_type(a=int, b=str, d=(int, None), e=float)
        def foo(a, b, c, d=None, **kwargs):
            pass
    """

    def check(
        self,
        type_spec: typing.Union[type, typing.Tuple[type, None]],
        arg_name: str,
        decorated_function: typing.Callable[..., typing.Any]
    ) -> typing.Callable[[typing.Any], None]:
        """ Return callable that checks function parameter for type validity. Checks parameter if it is
        instance of specified class or classes

        :param type_spec: type or list/tuple/set of types
        :param arg_name: function parameter name
        :param decorated_function: target function
        """

        def raise_exception(x_spec: str) -> None:
            exc_text = 'Argument "%s" for function "%s" has invalid type' % (
                arg_name, Verifier.function_name(decorated_function)
            )
            exc_text += ' (%s should be %s)' % (x_spec, type_spec)
            raise TypeError(exc_text)

        none_allowed = False
        type_only_spec = []

        if isinstance(type_spec, (tuple, list, set)):
            for single_type in type_spec:
                if single_type is None:
                    none_allowed = True
                elif isclass(single_type) is True:
                    type_only_spec.append(single_type)
                else:
                    raise RuntimeError(
                        'Invalid specification. Must be type or tuple/list/set of types'
                    )

            return lambda x: \
                None \
                if ((x is None and none_allowed) or (x is not None and isinstance(x, tuple(type_only_spec)))) \
                else \
                raise_exception(str(type(x)))

        elif isclass(type_spec):
            return lambda x: None if isinstance(x, type_spec) is True else \
                raise_exception(str(type(x)))
        else:
            raise RuntimeError('Invalid specification. Must be type or tuple/list/set of types')


class SubclassVerifier(Verifier):
    """ Verifier that is used for type verification. Checks parameter if it is class or subclass of
    specified class or classes. Specification accepts type or list/tuple/set of types

    Example:

        @verify_subclass(c=A, e=(A, D))
        def foo(a, b, c, d=None, **kwargs):
            pass
    """

    def check(
        self,
        type_spec: typing.Union[type, typing.Tuple[type, ...]],
        arg_name: str,
        decorated_function: typing.Callable[..., typing.Any]
    ) -> typing.Callable[[typing.Any], None]:
        """ Return callable that checks function parameter for class validity. Checks parameter if it is
        class or subclass of specified class or classes

        :param type_spec: type or tuple of types
        :param arg_name: function parameter name
        :param decorated_function: target function
        """

        def raise_exception(text_spec: str) -> None:
            exc_text = 'Argument "%s" for function "%s" has invalid type' % (
                arg_name, Verifier.function_name(decorated_function)
            )
            exc_text += ' (%s)' % text_spec
            raise TypeError(exc_text)

        if isinstance(type_spec, (tuple, list, set)):
            for single_type in type_spec:
                if (single_type is not None) and isclass(single_type) is False:
                    raise RuntimeError(
                        'Invalid specification. Must be type or tuple/list/set of types'
                    )
            if None in type_spec:
                type_spec = tuple(filter(lambda x: x is not None, type_spec))
                return lambda x: None if \
                    x is None or (isclass(x) is True and issubclass(x, type_spec) is True) else \
                    raise_exception(str(x))
            else:
                return lambda x: None if (isclass(x) is True and issubclass(x, type_spec) is True) else \
                    raise_exception(str(x))
        elif isclass(type_spec):
            return lambda x: None if (isclass(x) is True and issubclass(x, type_spec) is True) else \
                raise_exception(str(x))
        else:
            raise RuntimeError('Invalid specification. Must be type or tuple/list/set of types')


class ValueVerifier(Verifier):
    """ Verifier that is used for value verification. Checks parameter if its value passes specified restrictions.
    Specification accepts function or list/tuple/set of functions. Each function must accept one parameter and
    must return True or False if it passed restrictions or not.

    Example:

        @verify_value(a=(lambda x: x > 5, lambda x: x < 10))
        @verify_value(c=lambda x: x.a == 'foo', d=lambda x: x is None or x < 0)
        def foo(a, b, c, d=None, **kwargs):
        pass
    """

    def check(
        self,
        value_spec: typing.Union[
            typing.Callable[[typing.Any], bool], typing.Iterable[typing.Callable[[typing.Any], bool]]
        ],
        arg_name: str,
        decorated_function: typing.Callable[..., typing.Any]
    ) -> typing.Callable[[typing.Any], None]:
        """ Return callable that checks function parameter for value validity. Checks parameter if its value
        passes specified restrictions.

        :param value_spec: function or list/tuple/set of functions. Each function must accept one parameter and \
        must return True or False if it passed restrictions or not.
        :param arg_name: function parameter name
        :param decorated_function: target function
        """

        def raise_exception(text_spec: str) -> None:
            exc_text = 'Argument "%s" for function "%s" has invalid value' % (
                arg_name, Verifier.function_name(decorated_function)
            )
            exc_text += ' (%s)' % text_spec
            raise ValueError(exc_text)

        if isinstance(value_spec, (tuple, list, set)):

            for single_value in value_spec:
                if isfunction(single_value) is False:
                    raise RuntimeError(
                        'Invalid specification. Must be function or tuple/list/set of functions'
                    )

            def check(x: typing.Any) -> None:
                for f in value_spec:
                    if f(x) is not True:
                        raise_exception(str(x))

            return check

        elif isfunction(value_spec):
            return lambda x: None if value_spec(x) is True else raise_exception(str(x))
        else:
            raise RuntimeError('Invalid specification. Must be function or tuple/list/set of functions')


def verify_type(
    **type_kwargs: typing.Union[type, typing.Tuple[typing.Optional[type], ...]]
) -> typing.Callable[[GenericFunc[P, R]], GenericFunc[P, R]]:
    """ Shortcut for :class:`.TypeVerifier`

    :param type_kwargs: verifier specification. See :meth:`.TypeVerifier.check`
    :return: decorator (function)
    """
    return TypeVerifier().decorator(**type_kwargs)


def verify_subclass(
    **type_kwargs: typing.Union[type, typing.Tuple[type]]
) -> typing.Callable[[GenericFunc[P, R]], GenericFunc[P, R]]:
    """ Shortcut for :class:`.SubclassVerifier`

    :param type_kwargs: verifier specification. See :meth:`.SubclassVerifier.check`
    :return: decorator (function)
    """
    return SubclassVerifier().decorator(**type_kwargs)


def verify_value(
    **value_kwargs: typing.Union[
        typing.Callable[[typing.Any], bool],
        typing.Iterable[typing.Callable[[typing.Any], bool]]
    ]
) -> typing.Callable[[GenericFunc[P, R]], GenericFunc[P, R]]:
    """ Shortcut for :class:`.ValueVerifier`

    :param value_kwargs: verifier specification. See :meth:`.ValueVerifier.check`
    :return: decorator (function)
    """
    return ValueVerifier().decorator(**value_kwargs)



# -*- coding: utf-8 -*-
# pyknic/lib/signals/proxy.py
#
# Copyright (C) 2024 the pyknic authors and contributors
# <see AUTHORS file>
#
# This file is part of pyknic.
#
# pyknic is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# pyknic is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with pyknic.  If not, see <http://www.gnu.org/licenses/>.

import functools
import threading
import typing
import queue
import weakref

from pyknic.lib.signals.proto import SignalCallbackType, SignalProxyProto, SignalSourceProto, Signal
from pyknic.lib.signals.extra import CallbackWrapper

from pyknic.lib.tasks.proto import TaskProto


class SignalProxy(SignalProxyProto):
    """ This is a simple :class:`.SignalProxyProto` implementation
    """

    def __init__(self, wrapper_factory: typing.Optional[type[CallbackWrapper]] = None):
        """ Create a new proxy

        :param wrapper_factory: a class is used as a callback wrapper
        """
        SignalProxyProto.__init__(self)
        self.__callbacks: weakref.WeakValueDictionary[SignalCallbackType, SignalCallbackType] = \
            weakref.WeakValueDictionary()
        self.__wrapper_factory = wrapper_factory if wrapper_factory else CallbackWrapper

    def proxy(self, callback: SignalCallbackType) -> SignalCallbackType:
        """ The :meth:`.SignalProxyProto.proxy` method implementation
        """
        callback_wrapper = self.__wrapper_factory.wrapper(callback, weak_callback=True)
        self.__callbacks[callback_wrapper] = callback
        return callback_wrapper

    def discard_proxy(self, callback: SignalCallbackType) -> None:
        """ The :meth:`.SignalProxyProto.discard_proxy` method implementation
        """
        to_remove = []
        for wrapper, original_callback in self.__callbacks.items():
            if callback is original_callback:
                to_remove.append(wrapper)

        for i in to_remove:
            self.__callbacks.pop(i)


class QueueProxyStateError(Exception):
    """ This exception is raised when requested operation is not suitable in a current queue state
    """
    pass


class QueueCallbackException(Exception):
    """ This exception is raised when executed callback fails with exception
    """
    pass


class QueueProxy(SignalProxy, TaskProto):
    """ This "proxy" helps to execute "proxified" callbacks in a dedicated thread
    """

    class Item:
        """ This class represent a single callable item in a queue
        """

        def __init__(self, fn: typing.Callable[[], typing.Any]):
            """ Create new item for a queue
            """

            self.__fn = fn
            self.__wait_event = threading.Event()
            self.__result = None
            self.__raised_exception: typing.Optional[BaseException] = None

        def __call__(self) -> None:
            """ Execute a callback and check result
            """
            try:
                self.__result = self.__fn()
            except BaseException as e:
                self.__raised_exception = e
            self.__wait_event.set()

        def wait(self, timeout: typing.Union[int, float, None] = None) -> typing.Any:
            """ Wait for a callback result

            :param timeout: timeout with which a result will be awaited. If the timeout is None then wait forever
            """
            self.__wait_event.wait(timeout)
            if self.__raised_exception:
                raise QueueCallbackException('Callback completed with an exception') from self.__raised_exception
            return self.__result

    class Wrapper(CallbackWrapper):
        """ Alternate :class:`.CallbackWrapper` implementation that work in conjunction with the :class:`.QueueProxy`
        """

        def __init__(self, callback: SignalCallbackType, weak_callback: bool = False):
            """ Create a new wrapper

            :param callback: same as 'callback' in :meth:`.CallbackWrapper.__init__` method
            :param weak_callback: same as 'weak_callback' in :meth:`.CallbackWrapper.__init__` method
            """
            CallbackWrapper.__init__(self, callback, weak_callback=weak_callback)

            self.__queue: typing.Optional[queue.Queue[QueueProxy.Item]] = None
            self.__stop_event: typing.Optional[threading.Event] = None

        def setup_wrapper(
            self,
            queue_value: queue.Queue['QueueProxy.Item'],
            stop_event: threading.Event
        ) -> None:
            """ Setup this wrapper. Without this setup a wrapper will crash
            """
            assert(not self.__queue)
            assert(not self.__stop_event)

            self.__queue = queue_value
            self.__stop_event = stop_event

        def _pre_hook(
            self, callback: SignalCallbackType, source: SignalSourceProto, signal: Signal, value: typing.Any
        ) -> bool:
            """ This pre-hook always returns False (a callback will be skipped from follow-on execution) and submits
            the callback to a queue inside a :class:`.QueueProxy` object

            :param callback: a callback that is proxying
            :param source: signal origin
            :param signal: emitted signal
            :param value: a value of an emitted signal
            """
            assert(self.__queue)
            assert(self.__stop_event)

            if not self.__stop_event.is_set():
                self.__queue.put(QueueProxy.Item(functools.partial(
                    callback, source=source, signal=signal, value=value
                )))  # type: ignore[call-arg]  # mypy issue with functools.partial
            return False

    def __init__(self, flash_flush: bool = False) -> None:
        """ Create a new proxy

        :param flash_flush: if True then every submitted callback will be executed just before the stop event.
        Such callbacks may be omitted otherwise
        """
        SignalProxy.__init__(self, wrapper_factory=QueueProxy.Wrapper)
        TaskProto.__init__(self)

        self.__queue: queue.Queue[QueueProxy.Item] = queue.Queue()

        self.__start_once_lock = threading.Lock()
        self.__stop_once_lock = threading.Lock()
        self.__started_thread: typing.Optional[threading.Thread] = None
        self.__stop_event = threading.Event()
        self.__flash_flush = flash_flush

    def exec(
        self,
        fn: typing.Callable[[], typing.Any],
        blocking: bool = False,
        timeout: typing.Union[int, float, None] = None
    ) -> typing.Any:
        """ Execute a function inside a queue-thread

        :param fn: a callable to execute
        :param blocking: if True then the function result will be awaited
        :param timeout: timeout with which a result should be awaited. This parameter takes effect only when
        the "blocking" parameter is True
        """
        if self.__started_thread is None or self.__stop_event.is_set():
            raise QueueProxyStateError(
                'The "exec" method of the QueueProxy class may be called only if the QueueProxy is running'
            )

        queue_item = QueueProxy.Item(fn)
        self.__queue.put(queue_item)
        if blocking:
            return queue_item.wait(timeout)

    def proxy(self, callback: SignalCallbackType) -> SignalCallbackType:
        """ The :meth:`.SignalProxyProto.proxy` method implementation
        """
        callback_wrapper: QueueProxy.Wrapper = \
            SignalProxy.proxy(  # type: ignore[assignment]  # the SignalProxy.proxy method will return Wrapper
                self, callback
            )

        callback_wrapper.setup_wrapper(self.__queue, self.__stop_event)
        return callback_wrapper

    def __flush(self) -> None:
        """ Execute all the functions that have been stored in the queue
        """
        try:
            callback = self.__queue.get(False)
            while callback:
                # with the False value
                if not self.__flash_flush:
                    callback()
                self.__queue.task_done()
                callback = self.__queue.get(False)
        except queue.Empty:
            pass

    def is_running(self) -> bool:
        """ Return True if this queue is running and may accept items and return False otherwise
        """
        return self.__started_thread is not None and not self.__stop_event.is_set()

    def start(self) -> None:
        """ The :meth:`.TaskProto.start` method implementation
        """
        with self.__start_once_lock:
            if self.__started_thread is not None:
                raise QueueProxyStateError("Unable to start QueueProxy twice")
            self.__started_thread = threading.current_thread()

        while not self.__stop_event.is_set():
            next_callback = self.__queue.get()
            next_callback()
            self.__queue.task_done()

        self.__flush()

    def stop(self) -> None:
        """ The :meth:`.TaskProto.stop` method implementation
        """

        with self.__start_once_lock:
            if self.__started_thread is None:
                raise QueueProxyStateError("QueueProxy hasn't started yet")

        lock_acquire = self.__stop_once_lock.acquire(False)
        if lock_acquire:
            self.__stop_event.set()
            self.__queue.put(QueueProxy.Item(lambda: None))  # force the "self.__queue.get()" to return something
            self.__queue.join()
        else:
            raise QueueProxyStateError("QueueProxy can not br stopped twice")

    def is_inside(self) -> bool:
        """ This method helps to find whether current stack is inside started queue or not

        :return: return True if this function is called from the same thread as original :meth:`.QueueProxy.start`
        method is running
        """
        return self.__started_thread is threading.current_thread()



# -*- coding: utf-8 -*-
# pyknic/lib/fastapi/models/tg_bot_methods.py
#
# Copyright (C) 2024 the pyknic authors and contributors
# <see AUTHORS file>
#
# This file is part of pyknic.
#
# pyknic is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# pyknic is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with pyknic.  If not, see <http://www.gnu.org/licenses/>.


# TODO: autogenerate like this "https://github.com/AntonOvsyannikov/pure-teleapi/blob/main/apigen/__main__.py"

import pydantic
import typing

from pyknic.lib.fastapi.models.tg_bot_types import ReplyParameters, InlineKeyboardMarkup


# These models are based on https://core.telegram.org/bots/api#available-methods


class MethodSendMessageArgs(pydantic.BaseModel):
    # origin: https://core.telegram.org/bots/api#sendmessage
    chat_id: int | str
    text: str
    reply_parameters: ReplyParameters | None = None
    reply_markup: InlineKeyboardMarkup | None = None


class MethodSendMessage(MethodSendMessageArgs):
    # origin: https://core.telegram.org/bots/api#sendmessage
    method: typing.Literal['sendMessage'] = pydantic.Field(default='sendMessage', frozen=True)


class MethodAnswerCallbackQueryArgs(pydantic.BaseModel):
    # origin: https://core.telegram.org/bots/api#answercallbackquery
    callback_query_id: str
    text: str | None = None
    show_alert: bool | None = None


class MethodAnswerCallbackQuery(MethodAnswerCallbackQueryArgs):
    # origin: https://core.telegram.org/bots/api#answercallbackquery
    method: typing.Literal['answerCallbackQuery'] = pydantic.Field(default='answerCallbackQuery', frozen=True)



# -*- coding: utf-8 -*-

import asyncio
import gc
import pytest
import typing

if typing.TYPE_CHECKING:
    # noinspection PyUnresolvedReferences
    from conftest import CallbackRegistry, SignalsRegistry

from pyknic.lib.signals.proto import SignalSourceProto, Signal, SignalCallbackType
from pyknic.lib.signals.source import SignalSource
from pyknic.lib.signals.extra import BoundedCallback, CallbackWrapper, SignalResender, CallbacksHolder
from pyknic.lib.signals.extra import CustomizedCallback, SignalWaiter, AsyncWatchDog, ReceivedSignal

from fixtures.asyncio import pyknic_async_test


class TestBoundedCallback:

    def test(self, callbacks_registry: 'CallbackRegistry') -> None:

        class Source(SignalSource):
            signal1 = Signal()

        class A:

            def callback(self, source: SignalSourceProto, signal: Signal, value: typing.Any) -> None:
                callbacks_registry.callback('test_callback')()

        s = Source()

        b = BoundedCallback(A().callback)
        s.callback(Source.signal1, b)  # this is ok, but callback will not be executed
        # since 'A()' will be collected
        gc.collect()
        s.emit(Source.signal1)
        assert(callbacks_registry.calls('test_callback') == 0)

        a = A()
        b = BoundedCallback(a.callback)
        s.callback(Source.signal1, b)  # this is ok totally
        gc.collect()
        s.emit(Source.signal1)
        assert(callbacks_registry.calls('test_callback') == 1)

        del a
        gc.collect()
        s.emit(Source.signal1)
        assert(callbacks_registry.calls('test_callback') == 1)

    def test_private_method(self, callbacks_registry: 'CallbackRegistry') -> None:
        class Source(SignalSource):
            signal = Signal()

        class Callback:

            def __init__(self) -> None:
                self.__bounded_callback = BoundedCallback(self.__callback)

            def __callback(self, source: SignalSourceProto, signal: Signal, value: typing.Any) -> None:
                callbacks_registry.callback('test_callback')()

            def register(self, source: Source) -> None:
                source.callback(Source.signal, self.__bounded_callback)

        s = Source()
        c = Callback()
        c.register(s)
        s.emit(Source.signal)
        assert(callbacks_registry.calls('test_callback') == 1)

        del c
        gc.collect()
        s.emit(Source.signal)
        assert(callbacks_registry.calls('test_callback') == 1)


class TestCallbackWrapper:

    @pytest.mark.parametrize(
        "weak_ref, callback_called", [
            (None, True),
            (False, True),
            (True, False),
        ]
    )
    def test(self, signals_registry: 'SignalsRegistry', weak_ref: bool, callback_called: bool) -> None:

        class CallbackCls:

            def __call__(self, *args: typing.Any, **kwargs: typing.Any) -> None:
                signals_registry(*args, **kwargs)

        callback_obj = CallbackCls()
        callback_wrapper = CallbackWrapper.wrapper(callback_obj, weak_callback=weak_ref)

        callback_result = [(None, None, None)]

        callback_wrapper(None, None, None)  # type: ignore[arg-type]  # it's just a test
        assert(signals_registry.dump(True) == callback_result)

        del callback_obj
        gc.collect()
        callback_wrapper(None, None, None)  # type: ignore[arg-type]  # it's just a test
        if callback_called:
            assert (signals_registry.dump(True) == callback_result)
        else:
            assert(signals_registry.dump(True) == [])  # since signal_proxy is collected

    def test_hooks(self, signals_registry: 'SignalsRegistry') -> None:

        increment_value = 0

        class CallbackCls:

            def __call__(self, source: SignalSourceProto, signal: Signal, value: typing.Any) -> None:
                nonlocal increment_value
                increment_value += 1
                signals_registry(source, signal, increment_value)

        class CustomWrapper(CallbackWrapper):

            def _pre_hook(
                self, callback: SignalCallbackType, source: SignalSourceProto, signal: Signal, value: typing.Any
            ) -> bool:
                nonlocal increment_value
                increment_value += 1
                signals_registry(source, signal, {"callback": callback, "value": increment_value, "mode": "pre"})
                return True

            def _post_hook(
                self, callback: SignalCallbackType, source: SignalSourceProto, signal: Signal, value: typing.Any
            ) -> None:
                nonlocal increment_value
                increment_value += 1
                signals_registry(source, signal, {"callback": callback, "value": increment_value, "mode": "post"})

        callback_obj = CallbackCls()
        callback_wrapper = CustomWrapper.wrapper(callback_obj, weak_callback=True)

        callback_wrapper(None, None, None)  # type: ignore[arg-type]  # it's just a test
        assert(signals_registry.dump(True) == [
            (None, None, {"callback": callback_obj, "value": 1, "mode": "pre"}),
            (None, None, 2),
            (None, None, {"callback": callback_obj, "value": 3, "mode": "post"}),
        ])

        del callback_obj
        gc.collect()
        callback_wrapper(None, None, None)  # type: ignore[arg-type]  # it's just a test
        assert(signals_registry.dump(True) == [])

    def test_pre_hook(self, signals_registry: 'SignalsRegistry') -> None:

        class CallbackCls:

            def __call__(self, source: SignalSourceProto, signal: Signal, value: typing.Any) -> None:
                signals_registry(source, signal, value)

        class CustomWrapper(CallbackWrapper):

            def _pre_hook(
                self, callback: SignalCallbackType, source: SignalSourceProto, signal: Signal, value: typing.Any
            ) -> bool:
                signals_registry(source, signal, "pre")
                return False

            def _post_hook(
                self, callback: SignalCallbackType, source: SignalSourceProto, signal: Signal, value: typing.Any
            ) -> None:
                signals_registry(source, signal, "post")

        callback_obj = CallbackCls()
        callback_wrapper = CustomWrapper.wrapper(callback_obj, weak_callback=True)

        callback_wrapper(None, None, None)  # type: ignore[arg-type]  # it's just a test
        assert(signals_registry.dump(True) == [
            (None, None, "pre"),
        ])


class TestSignalResender:

    def test_plain(self, signals_registry: 'SignalsRegistry') -> None:
        class Source(SignalSource):
            signal = Signal()

        source1 = Source()
        source2 = Source()
        source2.callback(Source.signal, signals_registry)

        resender = SignalResender(source2)
        source1.callback(Source.signal, resender)
        source1.emit(Source.signal)  # source1 emitted, source2 re-emitted
        assert(signals_registry.dump(True) == [
            (source2, Source.signal, None),
        ])

    def test_different_signal(self, signals_registry: 'SignalsRegistry') -> None:

        class Source1(SignalSource):
            signal = Signal()

        class Source2(SignalSource):
            signal = Signal()

        source1 = Source1()
        source2 = Source2()
        source2.callback(Source2.signal, signals_registry)

        resender = SignalResender(source2, target_signal=Source2.signal)
        source1.callback(Source1.signal, resender)
        source1.emit(Source1.signal)  # source1 emitted, source2 re-emitted
        assert(signals_registry.dump(True) == [
            (source2, Source2.signal, None),
        ])

    def test_weak(self, signals_registry: 'SignalsRegistry') -> None:

        class Source(SignalSource):
            signal = Signal()

        source1 = Source()
        source2 = Source()
        source2.callback(Source.signal, signals_registry)

        resender = SignalResender(source2, Source.signal, weak_target=True)  # noqa: F841  # it must be so
        source1.callback(Source.signal, resender)
        del source2
        gc.collect()
        source1.emit(Source.signal)  # source1 emitted, source2 re-emitted
        assert(signals_registry.dump(True) == [])

    def test_value_converter(self, signals_registry: 'SignalsRegistry') -> None:
        class Source1(SignalSource):
            signal = Signal()

        class Source2(SignalSource):
            signal = Signal(SignalSourceProto)

        source1 = Source1()
        source2 = Source2()
        source2.callback(Source2.signal, signals_registry)

        def value_converter(source: SignalSourceProto, signal: Signal, value: typing.Any) -> typing.Any:
            return source

        resender = SignalResender(  # noqa: F841  # it must be so
            source2, Source2.signal, weak_target=True, value_converter=value_converter
        )
        source1.callback(Source1.signal, resender)

        source1.emit(Source1.signal)
        assert(signals_registry.dump(True) == [
            (source2, Source2.signal, source1),
        ])


class TestCallbacksHolder:

    def test(self, callbacks_registry: 'CallbackRegistry') -> None:
        holder = CallbacksHolder()

        class Source(SignalSource):
            signal = Signal()

        class Ref:
            pass

        reference = Ref()
        clbk = callbacks_registry.callback('test-callback')
        source = Source()

        source.callback(Source.signal, holder.keep_callback(lambda src, signal, value: clbk(), reference))
        assert(callbacks_registry.calls('test-callback') == 0)

        source.emit(Source.signal)
        assert(callbacks_registry.calls('test-callback') == 1)

        del reference
        gc.collect()
        source.emit(Source.signal)
        assert(callbacks_registry.calls('test-callback') == 1)


class TestCustomizedCallback:

    def test(self, callbacks_registry: 'CallbackRegistry') -> None:

        def custom_callback(
            source: SignalSourceProto,
            signal: Signal,
            value: typing.Any,
            optional_arg: typing.Optional[str] = None
        ) -> None:
            callbacks_registry.callback(optional_arg)()

        c_callbacks = CustomizedCallback()

        class Source(SignalSource):
            signal1 = Signal()
            signal2 = Signal()

        class Ref:
            pass

        reference1 = Ref()
        reference2 = Ref()
        source = Source()
        source.callback(Source.signal1, c_callbacks.customize(custom_callback, reference1, optional_arg='test-call1'))
        source.callback(Source.signal2, c_callbacks.customize(custom_callback, reference2, optional_arg='test-call2'))

        assert(callbacks_registry.calls('test-call1') == 0)
        assert(callbacks_registry.calls('test-call2') == 0)

        source.emit(Source.signal1)
        assert(callbacks_registry.calls('test-call1') == 1)
        assert(callbacks_registry.calls('test-call2') == 0)

        source.emit(Source.signal2)
        assert(callbacks_registry.calls('test-call1') == 1)
        assert(callbacks_registry.calls('test-call2') == 1)

        del reference1
        gc.collect()

        source.emit(Source.signal1)
        source.emit(Source.signal2)
        assert(callbacks_registry.calls('test-call1') == 1)
        assert(callbacks_registry.calls('test-call2') == 2)


class TestSignalWaiter:

    def test(self) -> None:

        class Source(SignalSource):
            signal = Signal(int)

        source = Source()

        waiter = SignalWaiter(source, Source.signal, timeout=0.1, value_matcher=lambda x: x == 5)
        assert(waiter.wait() is None)

        source.emit(Source.signal, 1)
        assert(waiter.wait() is None)

        source.emit(Source.signal, 5)
        assert(waiter.wait() == ReceivedSignal(source, Source.signal, 5))

    def test_context(self) -> None:

        class Source(SignalSource):
            signal = Signal(int)

        source = Source()
        with pytest.raises(TimeoutError):
            with SignalWaiter(source, Source.signal, timeout=0.1, value_matcher=lambda x: x == 5):
                source.emit(Source.signal, 7)

        with SignalWaiter(source, Source.signal, timeout=0.1, value_matcher=lambda x: x == 5):
            source.emit(Source.signal, 5)  # this is ok

    def test_multiple_signals(self) -> None:

        class Source(SignalSource):
            signal1 = Signal()
            signal2 = Signal()

        source1 = Source()
        source2 = Source()

        with SignalWaiter(source1, Source.signal1, timeout=0.1) as w:
            source2.callback(Source.signal2, w)
            source1.emit(Source.signal1)  # is ok

        with SignalWaiter(source1, Source.signal1, timeout=0.1) as w:
            source2.callback(Source.signal2, w)
            source2.emit(Source.signal2)  # is ok

        with pytest.raises(TimeoutError):
            with SignalWaiter(source1, Source.signal1, timeout=0.1) as w:
                source2.callback(Source.signal2, w)
                source1.emit(Source.signal2)  # will fail


class TestAsyncWatchDog:

    @pyknic_async_test
    async def test(self, module_event_loop: asyncio.AbstractEventLoop) -> None:

        class Source(SignalSource):
            signal1 = Signal(int)

        source1 = Source()

        async def emit_signal() -> None:
            await asyncio.sleep(0.1)
            source1.emit(Source.signal1, 10)

        watchdog = AsyncWatchDog(asyncio.get_event_loop(), source1, Source.signal1)

        watchdog_future, _ = await asyncio.gather(watchdog.wait(10), emit_signal())
        assert(watchdog_future.done() is True)
        assert(isinstance(watchdog_future.result(), ReceivedSignal) is True)
        assert(watchdog_future.result().source is source1)
        assert(watchdog_future.result().signal is Source.signal1)
        assert(watchdog_future.result().value == 10)

        # test that second trigger doesn't change result
        source1.emit(Source.signal1, 20)
        assert((await watchdog.watchdog_future()).result().value == 10)

    @pyknic_async_test
    async def test_wait_wo_result(self, module_event_loop: asyncio.AbstractEventLoop) -> None:

        class Source(SignalSource):
            signal1 = Signal(int)

        source1 = Source()
        watchdog = AsyncWatchDog(asyncio.get_event_loop(), source1, Source.signal1)
        watchdog_future = await watchdog.wait(0.5)
        assert(watchdog_future.done() is False)
        watchdog_future = await watchdog.wait(0.5)
        assert(watchdog_future.done() is False)

    @pyknic_async_test
    async def test_wait_before_watchdog(
        self, module_event_loop: asyncio.AbstractEventLoop
    ) -> None:

        class Source(SignalSource):
            signal1 = Signal(int)

        source1 = Source()
        watchdog = AsyncWatchDog(asyncio.get_event_loop(), source1, Source.signal1)
        source1.emit(Source.signal1, 10)

        watchdog_future = await watchdog.wait()
        assert(watchdog_future.done() is True)
        assert(watchdog_future.result().source is source1)
        assert(watchdog_future.result().signal is Source.signal1)
        assert(watchdog_future.result().value == 10)

    def test_exception(self) -> None:
        class Source(SignalSource):
            signal1 = Signal(int)

        source1 = Source()

        loop1 = asyncio.new_event_loop()
        loop2 = asyncio.new_event_loop()

        watchdog = AsyncWatchDog(loop1, source1, Source.signal1)
        with pytest.raises(RuntimeError):
            loop2.run_until_complete(watchdog.watchdog_future())



# -*- coding: utf-8 -*-
# pyknic/lib/io/crypto/kdf.py
#
# Copyright (C) 2017-2026 the pyknic authors and contributors
# <see AUTHORS file>
#
# This file is part of pyknic.
#
# pyknic is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# pyknic is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with pyknic.  If not, see <http://www.gnu.org/licenses/>.

import typing

from cryptography.hazmat.primitives.kdf.pbkdf2 import PBKDF2HMAC as C10yPBKDF2HMAC
from cryptography.hazmat.backends import default_backend

from pyknic.lib.verify import verify_value
from pyknic.lib.crypto.hash import __default_hashers_registry__
from pyknic.lib.crypto.random import random_bytes

# TODO: checkout Balloon and Argon2 algorithms. The first one (Balloon) is recommended by NIST but there is no support
#  from cryptography package and openssl utility (at least I didn't found). The second one (Argon2) is implemented
#  in cryptography and openssl but was not recommended by NIST


class PBKDF2:
    """ Wrapper for Cryptography io PBKDF2 implementation with NIST recommendation and HMAC is used as pseudorandom
    function

    NIST recommendation can be read here:
    http://nvlpubs.nist.gov/nistpubs/Legacy/SP/nistspecialpublication800-132.pdf (Recommendation for Password-Based
    Key Derivation)
    """

    __minimum_key_length__ = 20
    """ Minimum key length is specified at Appendix A, Section A.1 of Recommendation for Password-Based
    Key Derivation by NIST
    """

    __minimum_salt_length__ = 16
    """ Minimum salt length is specified at Section 5.1 of Recommendation for Password-Based Key Derivation by NIST
    """

    __default_salt_length__ = 64
    """ Salt length value used by default
    """

    __default_digest_generator_name__ = 'sha256'
    """ Hash-generator that is used by default
    """

    __minimum_iterations_count__ = 1000000
    """ 1000 is the minimum iteration Count is specified at Section 5.2 of Recommendation for Password-Based Key
    Derivation by NIST. But at the same time, this recommendation was published at 2010. So it is better to use
    more iterations like OWASP recommends
    """

    __default_iterations_count__ = 1000000
    """ The iteration count used by default
    """

    __default_derived_key_length__ = 16
    """ Length of derived key used by default
    """

    @verify_value(key=lambda x: len(x) >= PBKDF2.__minimum_key_length__)
    @verify_value(salt=lambda x: x is None or len(x) >= PBKDF2.__minimum_salt_length__)
    @verify_value(iterations_count=lambda x: x is None or x >= PBKDF2.__minimum_iterations_count__)
    @verify_value(hash_fn_name=lambda x: x is None or __default_hashers_registry__.has(x))
    def __init__(
        self,
        key: typing.Union[bytes],
        salt: typing.Optional[bytes] = None,
        derived_key_length: typing.Optional[int] = None,
        iterations_count: typing.Optional[int] = None,
        hash_fn_name: typing.Optional[str] = None
    ):
        """ Generate new key (derived key) with PBKDF2 algorithm

        :param key: password
        :param salt: salt to use (if no salt was specified, then it will be generated automatically)
        :param derived_key_length: length of byte-sequence to generate
        :param iterations_count: iteration count
        :param hash_fn_name: name of hash function to be used with HMAC
        """

        self.__salt = salt if salt is not None else self.generate_salt()

        if derived_key_length is None:
            derived_key_length = self.__default_derived_key_length__

        if iterations_count is None:
            iterations_count = self.__default_iterations_count__
        self.__iterations_count = iterations_count

        if hash_fn_name is None:
            hash_fn_name = self.__class__.__default_digest_generator_name__
        self.__hash_name = hash_fn_name

        hash_cls = __default_hashers_registry__.get(hash_fn_name)
        hash_obj = hash_cls()

        pbkdf2_obj = C10yPBKDF2HMAC(
            algorithm=hash_obj.c10y_algorithm(),
            length=derived_key_length,
            salt=self.__salt,
            iterations=self.__iterations_count,
            backend=default_backend()
        )

        self.__derived_key = pbkdf2_obj.derive(key)

    def salt(self) -> bytes:
        """ Return salt value (that was given in constructor or created automatically)
        """
        return self.__salt

    def iterations(self) -> int:
        """ Return number of iterations this KDF used
        """
        return self.__iterations_count

    def hash_name(self) -> str:
        """ Return name of hash function this KDF used
        """
        return self.__hash_name

    def derived_key(self) -> bytes:
        """ Return derived key
        """
        return self.__derived_key

    @classmethod
    @verify_value(length=lambda x: x is None or x >= PBKDF2.__minimum_salt_length__)
    def generate_salt(cls, length: typing.Optional[int] = None) -> bytes:
        """ Generate salt that can be used by this object

        :param length: target salt length
        """
        return random_bytes(length if length is not None else cls.__default_salt_length__)



# -*- coding: utf-8 -*-

import typing
import pathlib
import pytest

from pyknic.lib.config import Config
from pyknic.path import root_path


@pytest.fixture
def config_file() -> typing.Callable[[pathlib.Path], Config]:

    def return_config(path: pathlib.Path) -> Config:
        config = Config()
        with open(root_path / path) as f:
            config.merge_file(f)

        return config

    return return_config



# -*- coding: utf-8 -*-
import threading
import time

import pytest
import typing
import uuid

from datetime import datetime, timezone

if typing.TYPE_CHECKING:
    # noinspection PyUnresolvedReferences
    from conftest import SampleTasks

from pyknic.lib.registry import APIRegistry, register_api
from pyknic.lib.datalog.proto import DatalogProto
from pyknic.lib.datalog.datalog import Datalog
from pyknic.lib.signals.extra import SignalWaiter
from pyknic.lib.tasks.scheduler.chain_source import ChainedTaskState, ChainedTaskLogEntry, ChainedTasksSource
from pyknic.lib.tasks.scheduler.chain_source import ChainedTask
from pyknic.lib.tasks.scheduler.record import ScheduleRecord
from pyknic.lib.tasks.scheduler.scheduler import Scheduler
from pyknic.lib.tasks.scheduler.plain_sources import InstantTaskSource
from pyknic.lib.tasks.proto import TaskResult, SchedulerProto, SchedulerFeedback
from pyknic.lib.tasks.threaded_task import ThreadedTask


class SourceTestHelper:

    def __init__(self) -> None:
        self.scheduler = Scheduler()
        self.scheduler_thread = ThreadedTask(self.scheduler)
        self.api_registry = APIRegistry()

    def start(self) -> None:
        self.scheduler_thread.start()

    def stop(self) -> None:
        self.scheduler_thread.stop()
        self.scheduler_thread.wait()
        self.scheduler_thread.join()


@pytest.fixture
def source_helper(request: pytest.FixtureRequest) -> typing.Generator[SourceTestHelper, None, None]:
    helper = SourceTestHelper()
    helper.start()

    yield helper
    helper.stop()


class TestChainedTask:

    def test(self, source_helper: SourceTestHelper) -> None:
        class Task(ChainedTask):
            def start(self) -> None:
                pass

        source = ChainedTasksSource(registry=source_helper.api_registry)
        datalog = source.datalog()
        uid = uuid.uuid4()
        task = Task.create(datalog, 'sample-api-id', uid)
        assert(task.uid() == uid)
        assert(task.datalog() is datalog)
        assert(task.api_id() == 'sample-api-id')

    def test_wait_for_result(self, source_helper: SourceTestHelper) -> None:

        event = threading.Event()
        result_object = object()

        @register_api(source_helper.api_registry, 'test-task')
        class Task(ChainedTask):
            def start(self) -> None:
                nonlocal event, result_object  # noqa: F824
                time.sleep(0.5)
                event.set()

                self.save_result(result_object)

        source = ChainedTasksSource(registry=source_helper.api_registry)
        source_thread = ThreadedTask(source)
        source_thread.start()

        source_helper.scheduler.subscribe(source)
        source.execute('test-task')

        task = Task(source.datalog(), 'task', uuid.uuid4())
        assert(task.wait_for('test-task').result is result_object)  # type: ignore[union-attr]
        assert(source.wait_for(source.datalog(), 'test-task').result is result_object)  # type: ignore[union-attr]

        source_thread.stop()
        source_thread.wait()
        source_thread.join()

    def test_wait_for_none(self, source_helper: SourceTestHelper) -> None:

        event = threading.Event()

        @register_api(source_helper.api_registry, 'test-task')
        class Task(ChainedTask):
            def start(self) -> None:
                nonlocal event  # noqa: F824
                time.sleep(0.5)
                event.set()

        source = ChainedTasksSource(registry=source_helper.api_registry)
        source_thread = ThreadedTask(source)
        source_thread.start()

        source_helper.scheduler.subscribe(source)
        source.execute('test-task')

        task = Task(source.datalog(), 'task', uuid.uuid4())
        assert(task.wait_for('test-task') is None)
        assert(source.wait_for(source.datalog(), 'test-task') is None)

        source_thread.stop()
        source_thread.wait()
        source_thread.join()


class TestChainedTaskLogEntry:

    def test(self) -> None:
        api_id = 'super_class'
        uid = uuid.uuid4()
        state = ChainedTaskState.completed
        result = TaskResult()
        event_datetime = datetime.now(timezone.utc)

        entry = ChainedTaskLogEntry(api_id, uid, state, result)
        assert(entry.api_id is api_id)
        assert(entry.uid is uid)
        assert(entry.event_datetime >= event_datetime)
        assert(entry.state is state)
        assert(entry.result is result)

    def test_exceptions(self) -> None:
        api_id = 'super_class'
        uid = uuid.uuid4()
        state = ChainedTaskState.completed
        result = TaskResult()

        entry = ChainedTaskLogEntry(api_id, uid, state, result)

        with pytest.raises(AttributeError):
            entry.api_id = "bad_class"  # type: ignore[misc]  # it is a test =)


class TestChainedTasksSource:

    class Task(ChainedTask):

        def __init__(self, datalog: DatalogProto, api_id: str, uid: uuid.UUID) -> None:
            ChainedTask.__init__(self, datalog, api_id, uid)
            self.event = threading.Event()

        def start(self) -> None:
            self.event.wait()

        def stop(self) -> None:
            self.event.set()

    def test_plain_start(self, source_helper: SourceTestHelper) -> None:
        source = ChainedTasksSource(registry=source_helper.api_registry)
        source_thread = ThreadedTask(source)
        source_thread.start()

        @register_api(source_helper.api_registry, 'test-task')
        class TestClass(TestChainedTasksSource.Task):
            pass

        assert(source.started_task('test-task') is None)
        source_helper.scheduler.subscribe(source)
        source.execute('test-task')
        assert(source.started_task('test-task') is not None)

        source_thread.stop()
        source_thread.wait()
        source_thread.join()

    @pytest.mark.parametrize("repeats", range(100))
    def test_dependent_start(self, source_helper: SourceTestHelper, repeats: int) -> None:
        source = ChainedTasksSource(registry=source_helper.api_registry)
        source_thread = ThreadedTask(source)
        source_thread.start()

        @register_api(source_helper.api_registry, 'test-task1')
        class TestClass1(TestChainedTasksSource.Task):
            pass

        @register_api(source_helper.api_registry, 'test-task2')
        class TestClass2(TestChainedTasksSource.Task):

            @classmethod
            def dependencies(cls) -> typing.Optional[typing.Set[str]]:
                return {'test-task1'}

        @register_api(source_helper.api_registry, 'test-task3')
        class TestClass3(TestChainedTasksSource.Task):

            @classmethod
            def dependencies(cls) -> typing.Optional[typing.Set[str]]:
                return {'test-task2'}

        source_helper.scheduler.subscribe(source)
        source.execute('test-task1')
        source.execute('test-task3')

        assert(source.started_task('test-task1') is not None)
        assert(source.started_task('test-task2') is not None)
        assert(source.started_task('test-task3') is not None)

        source_thread.stop()
        source_thread.wait()
        source_thread.join()

    def test_skip_started_coverage(self, source_helper: SourceTestHelper) -> None:
        source = ChainedTasksSource(registry=source_helper.api_registry)
        source_thread = ThreadedTask(source)
        source_thread.start()

        @register_api(source_helper.api_registry, 'test-task1')
        class TestClass1(TestChainedTasksSource.Task):
            pass

        @register_api(source_helper.api_registry, 'test-task2')
        class TestClass2(TestChainedTasksSource.Task):

            @classmethod
            def dependencies(cls) -> typing.Optional[typing.Set[str]]:
                return {'test-task1'}

        @register_api(source_helper.api_registry, 'test-task3')
        class TestClass3(TestChainedTasksSource.Task):

            @classmethod
            def dependencies(cls) -> typing.Optional[typing.Set[str]]:
                return {'test-task2'}

        source_helper.scheduler.subscribe(source)
        source.execute('test-task1')

        source_thread.stop()
        source_thread.wait()
        source_thread.join()

    def test_exception(self, source_helper: SourceTestHelper) -> None:
        source = ChainedTasksSource(registry=source_helper.api_registry)
        source_thread = ThreadedTask(source)
        source_thread.start()

        with pytest.raises(ValueError):
            source.scheduler_feedback(source_helper.scheduler, SchedulerFeedback.source_unsubscribed)

        with pytest.raises(ValueError):
            source.scheduler_feedback(source_helper.scheduler, 1)  # type: ignore[arg-type]  # just a test

        @register_api(source_helper.api_registry, 'test-task1')
        class TestClass1(TestChainedTasksSource.Task):
            pass

        @register_api(source_helper.api_registry, 'test-task2')
        class TestClass2(TestChainedTasksSource.Task):

            @classmethod
            def dependencies(cls) -> typing.Optional[typing.Set[str]]:
                return {'test-task3'}

        @register_api(source_helper.api_registry, 'test-task3')
        class TestClass3(TestChainedTasksSource.Task):

            @classmethod
            def dependencies(cls) -> typing.Optional[typing.Set[str]]:
                return {'test-task3'}

        with pytest.raises(ValueError):
            source.execute('test-task1')

        source_helper.scheduler.subscribe(source)

        with pytest.raises(ValueError):
            source.scheduler_feedback(source_helper.scheduler, SchedulerFeedback.source_subscribed)

        source.execute('test-task1')

        with pytest.raises(ValueError):
            source.execute('test-task1')

        with pytest.raises(ValueError):
            source.execute('test-task2')

        with pytest.raises(ValueError):
            source.execute('test-task3')

        source_thread.stop()
        source_thread.wait()
        source_thread.join()

    def test_scheduler_exception(self, source_helper: SourceTestHelper, sample_tasks: 'SampleTasks') -> None:
        scheduler = Scheduler(1)
        scheduler_thread = ThreadedTask(scheduler)
        scheduler_thread.start()

        instant_source = InstantTaskSource()

        source = ChainedTasksSource(registry=source_helper.api_registry)
        source_thread = ThreadedTask(source)
        source_thread.start()

        @register_api(source_helper.api_registry, 'test-task1')
        class TestClass1(TestChainedTasksSource.Task):
            pass

        scheduler.subscribe(instant_source)
        instant_source.schedule_record(ScheduleRecord(sample_tasks.LongRunningTask(), source))  # scheduler is full now

        scheduler.subscribe(source)
        with pytest.raises(ValueError):
            source.execute('test-task1')

        source_thread.stop()
        source_thread.wait()
        source_thread.join()

        scheduler_thread.stop()
        scheduler_thread.wait()
        scheduler_thread.join()

    def test_datalog(self, source_helper: SourceTestHelper) -> None:
        source = ChainedTasksSource(registry=source_helper.api_registry)
        assert(isinstance(source.datalog(), DatalogProto) is True)

        datalog = Datalog()
        source = ChainedTasksSource(datalog=datalog, registry=source_helper.api_registry)
        assert(source.datalog() is datalog)

    def test_datalog_entries(self, source_helper: SourceTestHelper) -> None:
        source = ChainedTasksSource(registry=source_helper.api_registry)
        source_thread = ThreadedTask(source)
        source_thread.start()

        @register_api(source_helper.api_registry, 'test-task')
        class TestClass1(ChainedTask):
            event = threading.Event()

            def start(self) -> None:
                self.event.wait()

        assert(list(source.datalog().iterate()) == [])

        source_helper.scheduler.subscribe(source)
        waiter = SignalWaiter(source_helper.scheduler, SchedulerProto.scheduled_task_completed)
        source.execute('test-task')

        datalog = source.datalog()
        datalog_list = list(datalog.iterate())
        assert(len(datalog_list) == 1)
        assert(isinstance(datalog_list[0], ChainedTaskLogEntry) is True)
        assert(datalog_list[0].api_id == "test-task")
        assert(datalog_list[0].state == ChainedTaskState.started)
        assert(datalog_list[0].result is None)

        TestClass1.event.set()
        waiter.wait()

        source_thread.stop()
        source_thread.wait()
        source_thread.join()

        datalog_list = list(datalog.iterate())
        assert(len(datalog_list) == 2)



# -*- coding: utf-8 -*-
# pyknic/tasks/fastapi/tgbot_word_games.py
#
# Copyright (C) 2024 the pyknic authors and contributors
# <see AUTHORS file>
#
# This file is part of pyknic.
#
# pyknic is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# pyknic is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with pyknic.  If not, see <http://www.gnu.org/licenses/>.

# TODO: document the code
# TODO: write tests for the code

import enum
import typing

from pyknic.lib.config import Config
from pyknic.lib.registry import register_api
from pyknic.lib.fastapi.models.tg_bot_types import InlineKeyboardMarkup, InlineKeyboardButton, Update
from pyknic.lib.fastapi.models.tg_bot_methods import MethodAnswerCallbackQuery
from pyknic.lib.fastapi.apps_registry import __default_fastapi_apps_registry__
from pyknic.tasks.fastapi.word_games.cities import TGBotCityGame
from pyknic.lib.gettext import GetTextWrapper
from pyknic.lib.fastapi.base import TgBotBaseFastAPIApp, TgBotResponseType


@register_api(__default_fastapi_apps_registry__, "tgbot_word_games")
class TGBotWordGames(TgBotBaseFastAPIApp):

    @enum.unique
    class UserSessionMode(enum.Enum):
        null = enum.auto()
        city = enum.auto()

    def __init__(self, config: Config, translations: GetTextWrapper):
        TgBotBaseFastAPIApp.__init__(self, config, translations)

        self.__sessions: typing.Dict[int, TGBotWordGames.UserSessionMode] = dict()
        self.__city_game = TGBotCityGame(config, translations)

    async def process_command(self, msg: str, tg_update: Update) -> TgBotResponseType | None:
        assert(tg_update.message is not None)
        assert(tg_update.message.from_ is not None)

        lang = self.user_lang(tg_update.message.from_)

        if msg == "/reset":
            return self.reply(
                tg_update.message,
                lang.gettext('Choose a game'),
                reply_markup=InlineKeyboardMarkup(
                    inline_keyboard=[
                        [InlineKeyboardButton(text=lang.gettext('Cities (🇷🇺)'), callback_data='/reset-to-cities')]
                    ]
                )
            )
        return None

    async def process_message(self, tg_update: Update) -> TgBotResponseType | None:
        assert(tg_update.message is not None)
        assert(tg_update.message.from_ is not None)

        lang = self.user_lang(tg_update.message.from_)

        mode = self.__sessions.setdefault(tg_update.message.from_.id_, TGBotWordGames.UserSessionMode.null)
        if mode == TGBotWordGames.UserSessionMode.null:
            return self.reply(tg_update.message, lang.gettext('Start a game with the "/reset" command'))
        elif mode == TGBotWordGames.UserSessionMode.city:
            return await self.__city_game.process_request(tg_update)

    async def callback_query(self,  tg_update: Update) -> MethodAnswerCallbackQuery:
        assert(tg_update.callback_query is not None)

        if tg_update.callback_query.data is not None and tg_update.callback_query.from_ is not None:
            if tg_update.callback_query.data == "/reset-to-cities":
                self.__sessions[tg_update.callback_query.from_.id_] = TGBotWordGames.UserSessionMode.city
                await self.__city_game.reset(tg_update.callback_query.from_.id_, tg_update)

                lang = self.user_lang(tg_update.callback_query.from_)

                return MethodAnswerCallbackQuery(
                    callback_query_id=str(tg_update.callback_query.id_),
                    text=lang.gettext("Let's start! Name a city!"),
                    show_alert=True
                )

        return MethodAnswerCallbackQuery(callback_query_id=str(tg_update.callback_query.id_))

    @classmethod
    def bot_path(cls, config: Config) -> str:
        """Return configured path (as a part of url) for TG-bot."""
        return str(config["pyknic"]["fastapi"]["tgbot_word_games"]["url_path"])



# -*- coding: utf-8 -*-
# pyknic/lib/capability.py
#
# Copyright (C) 2018-2024 the pyknic authors and contributors
# <see AUTHORS file>
#
# This file is part of pyknic.
#
# pyknic is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# pyknic is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with pyknic.  If not, see <http://www.gnu.org/licenses/>.

import typing

from abc import ABCMeta
from inspect import isfunction

from pyknic.lib.verify import verify_value
from pyknic.lib.typing import GenericFunc, P, R


class CapabilityDescriptor:
    """ This class describes a single capability. Every capability (function) has its own descriptor
    """

    @verify_value(capability_name=lambda x: len(x) > 0)
    def __init__(self, capability_cls: type, capability_name: str):
        """ Create a new descriptor for some capability (function)

        :param capability_cls: class that defines a capability
        :param capability_name: function name
        """
        self.__cls = capability_cls
        self.__name = capability_name

    def cls(self) -> type:
        """ Return origin class
        """
        return self.__cls

    def name(self) -> str:
        """ Return capability name
        """
        return self.__name


def capability(f: GenericFunc[P, R]) -> GenericFunc[P, R]:
    """ Mark a decorated function as a capability

    :param f: function that is defined as a capability
    """
    f.__pyknic_capability__ = None  # type: ignore[attr-defined] # we are force the attribute to be
    return f


class CapabilitiesHolderMeta(ABCMeta):
    """ This metaclass sets capability descriptor for every newly marked capabilities
    """

    def __init__(cls, name: str, bases: typing.Tuple[type], namespace: typing.Dict[str, typing.Any]):
        """ Generate new class with this metaclass

        :param name: same as 'name' in :meth:`.ABCMeta.__init__` method
        :param bases: same as 'bases' in :meth:`.ABCMeta.__init__` method
        :param namespace: same as 'namespace' in :meth:`.ABCMeta.__init__` method
        """
        ABCMeta.__init__(cls, name, bases, namespace)

        for n in dir(cls):
            i = ABCMeta.__getattribute__(cls, n)
            if callable(i) and hasattr(i, '__pyknic_capability__'):
                if isfunction(i) is False:
                    raise TypeError(
                        'Only functions may be a "capability". "classmethod" or "staticmethod" '
                        'are not supported'
                    )
                if i.__pyknic_capability__ is None:
                    i.__pyknic_capability__ = CapabilityDescriptor(cls, i.__name__)


@verify_value(obj_capability_fn=lambda x: isinstance(x.__pyknic_capability__, CapabilityDescriptor))
def iscapable(obj: object, obj_capability_fn: typing.Callable[..., typing.Any]) -> bool:
    """ Check if the specified object (or type) is capable of something. I.e. check if a function from a descriptor
    has been overridden in the derived class. Return True if function has been overridden and False otherwise

    :param obj: object to check
    :param obj_capability_fn: capability descriptor
    """

    obj_capability = obj_capability_fn.__pyknic_capability__  # type: ignore[attr-defined] # metaclass/decorators issues

    obj_method = None
    if isinstance(obj, type):
        if issubclass(obj, obj_capability.cls()) is True:
            obj_method = getattr(obj, obj_capability.name())
    elif isinstance(obj, obj_capability.cls()) is True:
        obj_method = getattr(obj, obj_capability.name())

    if obj_method:
        return not hasattr(obj_method, '__pyknic_capability__')  # overridden methods will not have
        # the __pyknic_capability__ attribute

    return False


class CapabilitiesHolder(metaclass=CapabilitiesHolderMeta):
    """ Simple base class for defining capabilities. Besides that this class uses :class:`.CapabilitiesHolderMeta`
    metaclass already. This class overrides '__contains__' method, so 'in' operator may be used for checking
    whether a class has a capability
    """

    def __contains__(self, item: typing.Callable[..., typing.Any]) -> bool:
        """ Check whether this object has a capability

        :param item: capability to check
        """
        return iscapable(self, item)

    def append_capability(
        self, obj_capability_fn: typing.Callable[..., typing.Any], func: typing.Callable[..., typing.Any]
    ) -> None:
        """ Append capability implementation to the object

        :param obj_capability_fn: capability to implement
        :param func: capability implementation
        """
        cap_name = obj_capability_fn.__pyknic_capability__.name()  # type: ignore[attr-defined] # metaclass/decorators
        # issues

        if iscapable(self, obj_capability_fn):
            raise ValueError(f'An object already has the "{cap_name}" capability')

        setattr(self, cap_name, func)



# -*- coding: utf-8 -*-
# pyknic/lib/signals/proto.py
#
# Copyright (C) 2019-2024 the pyknic authors and contributors
# <see AUTHORS file>
#
# This file is part of pyknic.
#
# pyknic is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# pyknic is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with pyknic.  If not, see <http://www.gnu.org/licenses/>.

import typing

from abc import ABCMeta, abstractmethod
from inspect import isclass, isfunction, ismethod


class Signal:
    """ A signal that may be sent within pyknic.lib.signals methods
    """

    __pyknic_signal_name__: typing.Optional[str] = None

    def __init__(self, *checks: typing.Union[type, typing.Callable[[typing.Any], bool]]):
        """ Create a new (and unique) signal. Every object represent a unique signal

        :param checks: types or callables that signal value (that sent within a signal) must comply. Types are optional
        classes that signal value must be derived from. Callables (functions or methods) check values, each function
        """
        self.__check_types = tuple({x for x in checks if isclass(x)})
        self.__check_functions = tuple({x for x in checks if isfunction(x) or ismethod(x)})

    def check_value(self, value: typing.Any) -> None:
        """ Check value and raise exceptions (TypeError or ValueError) if value is invalid

        :param value: value that is checked whether it may or may not be sent with this signal

        :raises TypeError: if a value's type differ from types that are specified for this signal
        :raises ValueError: if a value's value does not comply with checks functions
        """
        if self.__check_types and isinstance(value, self.__check_types) is False:
            raise TypeError('Signal value is invalid (type mismatch)')
        for c in self.__check_functions:
            if not c(value):
                raise ValueError('Signal value is invalid (value mismatch)')

    def __hash__(self) -> int:
        """ Hash function in order to use this class as a dict key
        """
        return id(self)

    def __eq__(self, other: typing.Any) -> bool:
        """ Comparison function in order to use this class as a dict key
        """
        return id(other) == id(self)

    def __repr__(self) -> str:
        """ Because each signal is unique, it is much simpler to show its identifier in debug messages instead of an
        object's memory location
        """
        if self.__pyknic_signal_name__:
            return self.__pyknic_signal_name__
        return object.__repr__(self)


class SignalSourceProto(metaclass=ABCMeta):
    """ An entry class for an object that sends signals. Every callback is saved as a 'weak' reference. So in most
    cases in order to stop executing callback it is sufficient just to discard all callback's references
    """

    @abstractmethod
    def emit(self, signal: Signal, signal_value: typing.Any = None) -> None:
        """ Send a signal from this object and execute all linked callbacks within the same thread that called
        this function. Since number of callbacks are not limited it is better to use callbacks that postpone execution
        with some kind of "queue" rather than immediately execute a code. That queue may be an asyncio loop, thread pool
        or something similar

        :param signal: a signal (an object) to send
        :param signal_value: a signal argument that may be sent within a signal (this value should be checked first)
        """
        raise NotImplementedError('This method is abstract')

    @abstractmethod
    def callback(
        self,
        signal: Signal,
        callback: typing.Callable[['SignalSourceProto', Signal, typing.Any], None]
    ) -> None:
        """ Register a callback that will be executed when new signal is sent. Callbacks may be (and should be) saved
        as a weak reference so it is a callback responsibility to be alive and not to be collected by the mighty gc.
        If the same callback is registered twice it may be (or might be) called just once, no matter how many times
        it was registered

        :param signal: signal that will trigger a callback
        :param callback: callback that must be executed
        """
        raise NotImplementedError('This method is abstract')

    @abstractmethod
    def remove_callback(
        self,
        signal: Signal,
        callback: typing.Callable[['SignalSourceProto', Signal, typing.Any], None]
    ) -> None:
        """ Unregister the specified callback and prevent it to be executed when new signal is sent

        :param signal: signal that should be avoided by the specified callback
        :param callback: callback that should be unregistered
        """
        raise NotImplementedError('This method is abstract')


class SignalCallbackProto(metaclass=ABCMeta):
    """ An example of class that may receive signals (callback signature)
    """

    @abstractmethod
    def __call__(self, signal_source: SignalSourceProto, signal: Signal, signal_value: typing.Any = None) -> None:
        """ A callback that will be called when a signal is sent

        :param signal_source: origin of a signal
        :param signal: a signal that was sent
        :param signal_value: any argument that you want to pass with the specified signal. A specific signal
        may relay on this argument and may raise an exception if unsupported value is spotted
        """
        raise NotImplementedError('This method is abstract')


class UnknownSignalException(Exception):
    """ This exception may be raised if there was a request to amit an unknown signal. Usually it means that signal
    source is not able to send such signal.
    """
    pass


SignalCallbackType = typing.Callable[[SignalSourceProto, Signal, typing.Any], None]  # shortcut for the signal callback


class SignalProxyProto(metaclass=ABCMeta):
    """ This class works like a factory, which returns wrapper for callbacks
    """

    @abstractmethod
    def proxy(self, callback: SignalCallbackType) -> SignalCallbackType:
        """ Generate a wrapper for a callback and return it

        :param callback: a callback that should be "wrapped"
        :return: a wrapper
        """
        raise NotImplementedError('This method is abstract')

    @abstractmethod
    def discard_proxy(self, callback: SignalCallbackType) -> None:
        """ Disable previously returned wrappers that was generated by a callback

        :param callback: a callback which wrappers should be omitted
        """
        raise NotImplementedError('This method is abstract')



# -*- coding: utf-8 -*-
# pyknic/lib/fastapi/models/lobby.py
#
# Copyright (C) 2025-2026 the pyknic authors and contributors
# <see AUTHORS file>
#
# This file is part of pyknic.
#
# pyknic is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# pyknic is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with pyknic.  If not, see <http://www.gnu.org/licenses/>.

import typing

import pydantic

from pyknic.lib.fastapi.models.base import NullableModel
from pyknic.lib.fastapi.lobby_fingerprint import LobbyFingerprint
from pyknic.version import __version__


class LobbyCommandRequest(pydantic.BaseModel):
    """ This is a request for a command execution
    """
    model_config = pydantic.ConfigDict(extra='forbid')  # just to be sure that everything attributes are known
    client_version: str = __version__                   # a description of a client
    name: str                                           # a command to execute
    args: typing.Dict[str, typing.Any]                  # command's arguments (dictionary may be empty)


class LobbyStrFeedbackResult(pydantic.BaseModel):
    """Possible command result -- as a string"""
    model_config = pydantic.ConfigDict(extra='forbid')  # just to be sure that everything attributes are known
    server_version: str = __version__                   # a description of a server
    str_result: str                                     # a command result


class LobbyKeyValueFeedbackResult(pydantic.BaseModel):
    """Possible command result -- as a key-value pairs"""
    model_config = pydantic.ConfigDict(extra='forbid')  # just to be sure that everything attributes are known
    server_version: str = __version__                   # a description of a server
    kv_result: typing.Dict[str, typing.Any]             # a command result


class LobbyListValueFeedbackResult(pydantic.BaseModel):
    """Possible command result -- as a list of values"""
    model_config = pydantic.ConfigDict(extra='forbid')  # just to be sure that everything attributes are known
    server_version: str = __version__                   # a description of a server
    list_result: typing.List[typing.Any]                # a command result


# possible command results
LobbyCommandResult = typing.Union[
    NullableModel, LobbyStrFeedbackResult, LobbyKeyValueFeedbackResult, LobbyListValueFeedbackResult
]


class LobbyFingerprintModel(pydantic.BaseModel):
    """ A fingerprint of a lobby server
    """
    fingerprint: str = pydantic.Field(  # unique string. This is a string because of min/max length limitations
        min_length=LobbyFingerprint.serialized_length(),
        max_length=LobbyFingerprint.serialized_length(),
        validate_default=True
    )



# -*- coding: utf-8 -*-

import threading
import time
import typing

import pytest

if typing.TYPE_CHECKING:
    # noinspection PyUnresolvedReferences
    from conftest import SampleTasks, CallbackRegistry, SignalsRegistry, SignalWatcher

from datetime import datetime, timezone

from pyknic.lib.tasks.plain_task import PlainTask
from pyknic.lib.tasks.proto import TaskProto, ScheduledTaskPostponePolicy
from pyknic.lib.tasks.threaded_task import ThreadedTask

from pyknic.lib.tasks.scheduler.scheduler_executor import SchedulerExecutor


class TestSchedulerExecutor:

    class Task(TaskProto):

        def __init__(
            self,
            callback_registry: 'CallbackRegistry',
            callback_id: str
        ):
            TaskProto.__init__(self)
            self.event = threading.Event()
            self.__callback_registry = callback_registry
            self.__callback_id = callback_id

        def start(self) -> None:
            self.event.wait()
            self.__callback_registry.callback(self.__callback_id)()

    def test_plain(self, callbacks_registry: 'CallbackRegistry', sample_tasks: 'SampleTasks') -> None:
        executor = SchedulerExecutor()
        threaded_queue = ThreadedTask(executor.queue_proxy())
        record = sample_tasks.PlainRecord(PlainTask(callbacks_registry.callback('test-callback')))

        threaded_queue.start()
        executor.submit(record, blocking=True)

        executor.await_tasks()
        threaded_queue.stop()
        threaded_queue.join()

        assert(callbacks_registry.calls('test-callback') == 1)

    def test_scheduler_limits(self, callbacks_registry: 'CallbackRegistry', sample_tasks: 'SampleTasks') -> None:

        executor = SchedulerExecutor(2)
        threaded_queue = ThreadedTask(executor.queue_proxy())
        record1 = sample_tasks.PlainRecord(TestSchedulerExecutor.Task(callbacks_registry, 'test-callback'))
        record2 = sample_tasks.PlainRecord(TestSchedulerExecutor.Task(callbacks_registry, 'test-callback'))
        record3 = sample_tasks.PlainRecord(TestSchedulerExecutor.Task(callbacks_registry, 'test-callback'))
        record4 = sample_tasks.PlainRecord(TestSchedulerExecutor.Task(callbacks_registry, 'test-callback'))

        threaded_queue.start()

        time.sleep(1)

        executor.submit(record1, blocking=True)
        executor.submit(record2, blocking=True)
        executor.submit(record3, blocking=True)
        executor.submit(record4, blocking=True)

        record1.task().event.set()  # type: ignore[attr-defined]  # test simplification
        record2.task().event.set()  # type: ignore[attr-defined]  # test simplification
        record3.task().event.set()  # type: ignore[attr-defined]  # test simplification
        record4.task().event.set()  # type: ignore[attr-defined]  # test simplification

        for _ in range(100):
            if callbacks_registry.calls('test-callback') == 4:
                break
            time.sleep(0.1)
        else:
            assert(0)

        executor.await_tasks()
        threaded_queue.stop()
        threaded_queue.join()

    def test_outdated_ttl(self, signals_registry: 'SignalsRegistry', sample_tasks: 'SampleTasks') -> None:
        executor = SchedulerExecutor(1)
        threaded_queue = ThreadedTask(executor.queue_proxy())
        executor.callback(SchedulerExecutor.scheduled_task_expired, signals_registry)
        threaded_queue.start()

        old_record = sample_tasks.PlainRecord(sample_tasks.LongRunningTask(), ttl=1)
        executor.submit(old_record, blocking=True)

        assert(signals_registry.dump(True) == [
            (executor, SchedulerExecutor.scheduled_task_expired, old_record),
        ])

        executor.await_tasks()
        threaded_queue.stop()
        threaded_queue.join()

    def test_expired_ttl(self, signals_registry: 'SignalsRegistry', sample_tasks: 'SampleTasks') -> None:
        executor = SchedulerExecutor(1)
        threaded_queue = ThreadedTask(executor.queue_proxy())
        executor.callback(SchedulerExecutor.scheduled_task_expired, signals_registry)
        threaded_queue.start()

        long_run_record = sample_tasks.PlainRecord(sample_tasks.LongRunningTask())
        executor.submit(long_run_record, blocking=True)

        outdated_record = sample_tasks.PlainRecord(
            sample_tasks.LongRunningTask(), ttl=(datetime.now(timezone.utc).timestamp() + 3)
        )
        executor.submit(outdated_record, blocking=True)

        assert(signals_registry.dump(True) == [])
        time.sleep(5)

        executor.stop_running_tasks()
        executor.await_tasks()
        threaded_queue.stop()
        threaded_queue.join()

        assert(signals_registry.dump(True) == [
            (executor, SchedulerExecutor.scheduled_task_expired, outdated_record),
        ])

    def test_group_id_postpone(self, signals_registry: 'SignalsRegistry', sample_tasks: 'SampleTasks') -> None:
        executor = SchedulerExecutor(2)
        threaded_queue = ThreadedTask(executor.queue_proxy())
        executor.callback(SchedulerExecutor.scheduled_task_postponed, signals_registry)
        threaded_queue.start()

        first_record = sample_tasks.PlainRecord(sample_tasks.LongRunningTask(), group_id='test-group')
        executor.submit(first_record, blocking=True)

        assert(signals_registry.dump(True) == [])

        second_record = sample_tasks.PlainRecord(
            sample_tasks.LongRunningTask(), group_id='test-group', simultaneous_runs=1
        )
        executor.submit(second_record, blocking=True)

        assert(signals_registry.dump(True) == [
            (executor, SchedulerExecutor.scheduled_task_postponed, second_record)
        ])

        executor.cancel_postponed_tasks()
        executor.stop_running_tasks()
        executor.await_tasks()
        threaded_queue.stop()
        threaded_queue.join()

        assert(signals_registry.dump(True) == [])

    def test_postpone(self, signals_registry: 'SignalsRegistry', sample_tasks: 'SampleTasks') -> None:
        executor = SchedulerExecutor(1)
        threaded_queue = ThreadedTask(executor.queue_proxy())
        executor.callback(SchedulerExecutor.scheduled_task_postponed, signals_registry)
        threaded_queue.start()

        first_record = sample_tasks.PlainRecord(sample_tasks.LongRunningTask())
        executor.submit(first_record, blocking=True)

        assert(signals_registry.dump(True) == [])

        second_record = sample_tasks.PlainRecord(sample_tasks.LongRunningTask())
        executor.submit(second_record, blocking=True)

        assert(signals_registry.dump(True) == [
            (executor, SchedulerExecutor.scheduled_task_postponed, second_record)
        ])

        executor.cancel_postponed_tasks()
        executor.stop_running_tasks()
        executor.await_tasks()
        threaded_queue.stop()
        threaded_queue.join()

        assert(signals_registry.dump(True) == [])

    def test_complete_signal(self, signals_registry: 'SignalsRegistry', sample_tasks: 'SampleTasks') -> None:
        executor = SchedulerExecutor()
        threaded_queue = ThreadedTask(executor.queue_proxy())
        executor.callback(SchedulerExecutor.scheduled_task_completed, signals_registry)
        threaded_queue.start()

        record = sample_tasks.PlainRecord(sample_tasks.LongRunningTask())
        executor.submit(record, blocking=True)

        assert(signals_registry.dump(True) == [])

        executor.stop_running_tasks()
        executor.await_tasks()
        threaded_queue.stop()
        threaded_queue.join()

        assert(signals_registry.dump(True) == [
            (executor, SchedulerExecutor.scheduled_task_completed, record),
        ])

    def test_dropped(
        self, signals_registry: 'SignalsRegistry', sample_tasks: 'SampleTasks'
    ) -> None:
        executor = SchedulerExecutor(1)
        threaded_queue = ThreadedTask(executor.queue_proxy())
        executor.callback(SchedulerExecutor.scheduled_task_dropped, signals_registry)
        threaded_queue.start()

        first_record = sample_tasks.PlainRecord(sample_tasks.LongRunningTask())
        executor.submit(first_record, blocking=True)

        assert(signals_registry.dump(True) == [])

        second_record = sample_tasks.PlainRecord(
            sample_tasks.LongRunningTask(), postpone_policy=ScheduledTaskPostponePolicy.drop
        )
        executor.submit(second_record, blocking=True)

        assert(signals_registry.dump(True) == [
            (executor, SchedulerExecutor.scheduled_task_dropped, second_record)
        ])

        executor.stop_running_tasks()
        executor.await_tasks()
        threaded_queue.stop()
        threaded_queue.join()

        assert(signals_registry.dump(True) == [])

    def test_keep_first_signal(self, signal_watcher: 'SignalWatcher', sample_tasks: 'SampleTasks') -> None:
        executor = SchedulerExecutor(1)
        threaded_queue = ThreadedTask(executor.queue_proxy())
        executor.callback(SchedulerExecutor.scheduled_task_dropped, signal_watcher)
        threaded_queue.start()

        first_record = sample_tasks.PlainRecord(sample_tasks.LongRunningTask())
        executor.submit(first_record, blocking=True)

        second_record = sample_tasks.PlainRecord(sample_tasks.LongRunningTask(), group_id='test-group1')
        executor.submit(second_record, blocking=True)

        third_record = sample_tasks.PlainRecord(
            sample_tasks.LongRunningTask(),
            postpone_policy=ScheduledTaskPostponePolicy.keep_first,
            group_id='test-group1'
        )
        executor.submit(third_record, blocking=True)

        signal_watcher.wait(100)

        executor.cancel_postponed_tasks()
        executor.stop_running_tasks()
        executor.await_tasks()
        threaded_queue.stop()
        threaded_queue.join()

    def test_exception(self, sample_tasks: 'SampleTasks') -> None:
        executor = SchedulerExecutor()
        threaded_queue = ThreadedTask(executor.queue_proxy())
        threaded_queue.start()

        task = sample_tasks.LongRunningTask()
        first_record = sample_tasks.PlainRecord(task)
        executor.submit(first_record, blocking=True)

        second_record = sample_tasks.PlainRecord(task)
        with pytest.raises(ValueError):
            executor.submit(second_record, blocking=True)

        executor.stop_running_tasks()
        executor.await_tasks()
        threaded_queue.stop()
        threaded_queue.join()

    @pytest.mark.parametrize(
        "task_args", [
            {"stop_method": True, "terminate_method": False},
            {"stop_method": False, "terminate_method": True}
        ]
    )
    def test_stop_running_tasks(self, sample_tasks: 'SampleTasks', task_args: typing.Dict[str, bool]) -> None:
        executor = SchedulerExecutor(1)
        threaded_queue = ThreadedTask(executor.queue_proxy())
        threaded_queue.start()

        first_task = sample_tasks.LongRunningTask(**task_args)
        executor.submit(sample_tasks.PlainRecord(first_task), blocking=True)

        second_task = sample_tasks.LongRunningTask(**task_args)
        executor.submit(sample_tasks.PlainRecord(second_task), blocking=True)

        assert(executor.running_tasks() == (first_task, ))
        assert(executor.pending_tasks() == (second_task, ))

        executor.cancel_postponed_tasks()
        executor.stop_running_tasks()
        executor.await_tasks()
        threaded_queue.stop()
        threaded_queue.join()

    def test_await_exception(self, sample_tasks: 'SampleTasks') -> None:
        executor = SchedulerExecutor()
        threaded_queue = ThreadedTask(executor.queue_proxy())
        threaded_queue.start()

        record = sample_tasks.PlainRecord(sample_tasks.LongRunningTask())
        executor.submit(record, blocking=True)

        import time
        time.sleep(1)

        with pytest.raises(TimeoutError):
            executor.await_tasks(1)

        executor.stop_running_tasks()
        executor.await_tasks()
        threaded_queue.stop()
        threaded_queue.join()

    def test_corrupted_task(self, sample_tasks: 'SampleTasks') -> None:

        class Task(TaskProto):
            def start(self) -> None:
                raise ValueError('!')

        executor = SchedulerExecutor()

        threaded_queue = ThreadedTask(executor.queue_proxy())
        threaded_queue.start()

        task = Task()
        record = sample_tasks.PlainRecord(task)
        executor.submit(record, blocking=True)

        executor.stop_running_tasks()
        executor.await_tasks()
        threaded_queue.stop()
        threaded_queue.join()
